from typing import Union, Iterable

from .._content_provider import ContentProviderLayer
from ..search import SearchViews

from ._symbol_type import SymbolTypes, SYMBOL_TYPE_VALUES
from ._asset_state import AssetState
from ._asset_class import AssetClass
from ._country_code import CountryCode


search_all_category_by_asset_class = {
    AssetClass.COMMODITIES: "Commodities",
    AssetClass.EQUITY_OR_INDEX_OPTIONS: "Options",
    AssetClass.BOND_AND_STIR_FUTURES_AND_OPTIONS: "Exchange-Traded Rates",
    AssetClass.EQUITIES: "Equities",
    AssetClass.EQUITY_INDEX_FUTURES: "Futures",
    AssetClass.FUNDS: "Funds",
    AssetClass.BONDS: "Bond Pricing",
    AssetClass.FX_AND_MONEY: "FX & Money",
}

rcsasset_category_genealogy_by_asset_class = {
    AssetClass.WARRANTS: "A:AA",
    AssetClass.CERTIFICATES: "A:6N",
    AssetClass.INDICES: "I:17",
    AssetClass.RESERVE_CONVERTIBLE: "A:LE",
    AssetClass.MINI_FUTURE: "A:P6",
}


def _transform_to_string(values: Iterable, category: dict) -> str:
    string_of_values = ""
    for value in values:
        request_value = category[value]
        string_of_values = f"{string_of_values} '{request_value}'"
    return string_of_values


def _create_asset_class_request_strings(asset_class: list) -> tuple:
    search_all_category_values = filter(
        lambda x: x in search_all_category_by_asset_class, asset_class
    )
    rcs_asset_category_values = filter(
        lambda x: x in rcsasset_category_genealogy_by_asset_class, asset_class
    )

    search_all_category_string_values = _transform_to_string(
        search_all_category_values, search_all_category_by_asset_class
    )

    search_all_rcs_asset_category_string_values = _transform_to_string(
        rcs_asset_category_values, rcsasset_category_genealogy_by_asset_class
    )

    search_all_category_string = ""
    rcs_asset_category_string = ""

    if search_all_category_string_values:
        search_all_category_string = (
            f"SearchAllCategoryv3 in ({search_all_category_string_values})"
        )

    if search_all_rcs_asset_category_string_values:
        rcs_asset_category_string = f"RCSAssetCategoryGenealogy in ({search_all_rcs_asset_category_string_values})"

    return search_all_category_string, rcs_asset_category_string


def _prepare_filter(
    asset_state: AssetState, asset_class: Union[list, AssetClass]
) -> str:
    asset_state = asset_state or AssetState.ACTIVE

    if asset_state is AssetState.ACTIVE:
        ret_val = "AssetState eq 'AC'"
    else:
        ret_val = "(AssetState ne 'AC' and AssetState ne null)"

    if asset_class and not isinstance(asset_class, list):
        asset_class = [asset_class]

    if asset_class:
        search_all_category, rcs_asset_category = _create_asset_class_request_strings(
            asset_class
        )

        if search_all_category and rcs_asset_category:
            ret_val = f"{ret_val} and ({search_all_category} or {rcs_asset_category})"
        else:
            ret_val = f"{ret_val} and ({search_all_category}{rcs_asset_category})"

    return ret_val


class SymbologyContentProviderLayer(ContentProviderLayer):
    def __init__(self, **kwargs):
        from .._content_type import ContentType

        asset_state = kwargs.get("asset_state")
        asset_class = kwargs.get("asset_class")
        filter_ = _prepare_filter(asset_state, asset_class)

        preferred_country_code = kwargs.get("preferred_country_code")
        if preferred_country_code:
            preferred_country_code = f"RCSExchangeCountry eq '{CountryCode.convert_to_str(preferred_country_code)}'"

        from_symbol_type = kwargs.get("from_symbol_type")
        if from_symbol_type and from_symbol_type != "_AllUnique":
            from_symbol_type = SymbolTypes.convert_to_str(from_symbol_type)

        select = ["DocumentTitle"]
        to_symbol_types = kwargs.get("to_symbol_types")
        if to_symbol_types is SYMBOL_TYPE_VALUES:
            select += to_symbol_types
        elif isinstance(to_symbol_types, list):
            select += list(map(SymbolTypes.convert_to_str, to_symbol_types))
        elif isinstance(to_symbol_types, str):
            select.append(to_symbol_types)
        elif isinstance(to_symbol_types, SymbolTypes):
            select.append(to_symbol_types.value)

        select = ",".join(select)

        symbols = kwargs.get("symbols")
        if isinstance(symbols, list):
            symbols = ",".join(symbols)

        super().__init__(
            content_type=ContentType.DISCOVERY_LOOKUP,
            **{
                "view": SearchViews.SEARCH_ALL,
                "terms": symbols,
                "scope": from_symbol_type,
                "select": select,
                "extended_params": kwargs.get("extended_params"),
                "closure": kwargs.get("closure"),
                "filter": filter_,
                "boost": preferred_country_code,
            },
        )



# coding: utf-8

from .grant import Grant


class GrantPassword(Grant):
    def __init__(self, *args, **kwargs):
        """

        Parameters
        ----------
        *args: iterable
            args[0] is username, args[1] is password, args[2] is token scope (default: "trapi")
        **kwargs
            username: str, password: str

        Examples
        --------
        >>> import refinitiv.data as rdp
        >>> rdp.session.platform.GrantPassword(username="username", password="password", token_scope="token_scope")
        or
        >>> import refinitiv.data as rdp
        >>> rdp.session.platform.GrantPassword("username", "password", "token_scope")

        """
        super(GrantPassword, self).__init__(*args, **kwargs)

        if len(args) == 2:
            self._password = args[1]
        elif "password" in kwargs:
            self._password = kwargs["password"]

        self._token_scope = ""

        if len(args) == 3:
            self._token_scope = args[2]
        else:
            self._token_scope = kwargs.get("token_scope", "trapi")

        default_scope = "trapi"
        self._token_scope = self._token_scope or default_scope

    def get_password(self):
        return self._password

    def password(self, value):
        self._password = value
        return self

    def get_token_scope(self):
        return self._token_scope

    def with_scope(self, token_scope):
        if token_scope:
            self._token_scope = token_scope
        return self

    def is_valid(self):
        return self._password and self._username



from enum import Enum


class StatTypes(Enum):
    INVESTOR_TYPE = 1
    INVESTMENT_STYLE = 2
    REGION = 3
    ROTATION = 4
    COUNTRY = 5
    METRO_AREA = 6
    INVESTOR_TYPE_PARENT = 7
    INVESTOR_STYLE_PARENT = 8


class Frequency(Enum):
    QUARTERLY = "Q"
    MONTHLY = "M"


class SortOrder(Enum):
    ASCENDING = "asc"
    DESCENDING = "desc"



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import config as ext_config_mod
    from ._core.session import Session


class ConfigManager:
    """
    Temporary class
    When all content is moved to the data provider style,
    this class and module will be deleted
    """

    def get_api_config(
        self,
        api_name: str,
        session: "Session",
    ) -> "ext_config_mod.ConfigurationSet":
        config = session.config
        api_config = config.get(f"apis.{api_name}")
        return api_config


config_mgr: ConfigManager = ConfigManager()



import os
import urllib.parse

from ..._tools import parse_url, urljoin, cfs_datetime_adapter

forbidden_symbols = ':/|\\?*<>"'


def _convert_name(name: str) -> str:
    """
    Convert names from snake_case to camel case
    with first letter in lower case

    Parameters
    ----------
    name : str

    Examples
    --------
    >>> _convert_name("fileset_id")
    ... 'filesetId'
    """
    if "_" not in name:
        return name

    return "".join(
        word if i == 0 else word.title() for i, word in enumerate(name.split("_"))
    )


def _get_query_params(**kwargs):
    _query_parameters = []
    for (key, value) in kwargs.items():
        if value is not None and not key.startswith("_"):
            _query_parameters.append((_convert_name(key), value))
    return _query_parameters


def _get_url(config, endpoint):
    base_url = config.get_str("url")
    endpoint_url = config.get_str(f"endpoints.{endpoint}")
    _url = urljoin(base_url, endpoint_url)
    return _url


def _get_query_parameter(param, url):
    # url -> "site.com/package?par1=val1&par2=val2
    query = parse_url(url).query
    # query -> "par1=val1&par2=val2
    query_params = dict(urllib.parse.parse_qsl(query))
    # query_params -> {"par1": "val1", "par2": "val2"}
    return query_params.get(param)


def _convert_date_time(value):
    if value is None:
        return None
    return cfs_datetime_adapter.get_str(value)


def path_join(*args):
    path = os.path.join(*args)
    return os.path.normpath(path)


def create_dirs_if_no_exists(folder):
    if not folder:
        return

    if not os.path.exists(folder):
        os.makedirs(folder)


def get_correct_filename(filename):
    result = filename
    for symbol in forbidden_symbols:
        result = result.replace(symbol, "")
    return result


def remove_one_ext(filename_ext_ext):
    filename_ext, *_ = filename_ext_ext.rsplit(".", 1)
    return filename_ext


def remove_ext(filename_ext):
    return remove_one_ext(filename_ext)



# coding: utf8

from typing import Optional

from ._enums import RepoCurveType
from .._instrument_pricing_parameters import InstrumentPricingParameters


class PricingParameters(InstrumentPricingParameters):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    repo_curve_type : RepoCurveType, optional
        Curve used to compute the repo rate.
        If no curve can be found, the rate is computed using a deposit curve.
    market_data_date : str, optional
        The market data date for pricing.
        By default, the market_data_date date is the valuation_date or Today
    valuation_date : str, optional
        The valuation date for pricing. If not set the valuation date is equal to
        market_data_date or Today. For assets that contains a settlementConvention,
        the default valuation date  is equal to the settlementdate of the Asset that
        is usually the TradeDate+SettlementConvention.

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> rdf.repo.PricingParameters(market_data_date="2019-11-25")
    """

    def __init__(
        self,
        repo_curve_type: Optional[RepoCurveType] = None,
        market_data_date: Optional[str] = None,
        valuation_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.repo_curve_type = repo_curve_type
        self.market_data_date = market_data_date
        self.valuation_date = valuation_date

    @property
    def repo_curve_type(self):
        """
        Curve used to compute the repo rate. it can be computed using following methods:
            - RepoCurve : rate is computed by interpolating a repo curve.
            - DepositCurve : rate is computed by interpolating a deposit curve.
            - LiborFixing : rate is computed by interpolating libor rates.
        If no curve can be found, the rate is computed using a deposit curve.
        :return: enum RepoCurveType
        """
        return self._get_enum_parameter(RepoCurveType, "repoCurveType")

    @repo_curve_type.setter
    def repo_curve_type(self, value):
        self._set_enum_parameter(RepoCurveType, "repoCurveType", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing.
        By default, the market_data_date date is the valuation_date or Today
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing. If not set the valuation date is equal to
        market_data_date or Today. For assets that contains a settlementConvention, the
        default valuation date  is equal to the settlementdate of the Asset that is
        usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.212840

__all__ = ["RoundingType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class RoundingType(Enum):
    CEIL = "Ceil"
    DEFAULT = "Default"
    DOWN = "Down"
    FACE_DOWN = "FaceDown"
    FACE_NEAR = "FaceNear"
    FACE_UP = "FaceUp"
    FLOOR = "Floor"
    NEAR = "Near"
    UP = "Up"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(RoundingType, _ROUNDING_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_ROUNDING_TYPE_VALUES_IN_LOWER_BY_ROUNDING_TYPE, some)


_ROUNDING_TYPE_VALUES = tuple(t.value for t in RoundingType)
_ROUNDING_TYPE_VALUES_IN_LOWER_BY_ROUNDING_TYPE = {
    name.lower(): item for name, item in RoundingType.__members__.items()
}



# coding: utf8

__all__ = ["BusinessDayConvention"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class BusinessDayConvention(Enum):
    """
    The method to adjust dates to a working day.

    The possible values are:
        - BbswModifiedFollowing - dates are adjusted to the next business day convention
            unless is it goes into the next month, or crosses mid-month (15th).
            In this case, the Previous Business Day convention is used.
        - EveryThirdWednesday - dates are adjusted to the next every third Wednesday.
        - ModifiedFollowing - dates are adjusted to the next business day convention
            unless is it goes into the next month. In this case, the Previous Business
            Day convention is used.
        - NextBusinessDay - dates are moved to the following working day.
        - NoMoving - dates are not adjusted.
        - PreviousBusinessDay - dates are moved to the preceding working day.
    """

    BBSW_MODIFIED_FOLLOWING = "BbswModifiedFollowing"
    EVERY_THIRD_WEDNESDAY = "EveryThirdWednesday"
    MODIFIED_FOLLOWING = "ModifiedFollowing"
    NEXT_BUSINESS_DAY = "NextBusinessDay"
    NO_MOVING = "NoMoving"
    PREVIOUS_BUSINESS_DAY = "PreviousBusinessDay"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(BusinessDayConvention, _CONVENTION_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_CONVENTION_VALUES_IN_LOWER_BY_CONVENTION, some)


_CONVENTION_VALUES = tuple(t.value for t in BusinessDayConvention)
_CONVENTION_VALUES_IN_LOWER_BY_CONVENTION = {
    name.lower(): item for name, item in BusinessDayConvention.__members__.items()
}



import re
from typing import Dict

from ._data_grid_type import use_field_names_in_headers_arg_parser
from ..._tools import universe_arg_parser, fields_arg_parser, ADC_TR_PATTERN
from ...delivery._data._data_provider import (
    DataProvider,
    RequestFactory,
    ResponseFactory,
    ContentValidator,
    request_method_value,
    ValidatorContainer,
)
from ...delivery._data._endpoint_data import RequestMethod

# --------------------------------------------------------------------------------------
#   Response
# --------------------------------------------------------------------------------------
response_errors = {
    "default": "{error_message} Requested universes: {universes}. Requested fields: {"
    "fields}",
    412: "Unable to resolve all requested identifiers in {universes}.",
    218: "Unable to resolve all requested fields in {fields}. The formula must "
    "contain at least one field or function.",
}


class DataGridRDPResponseFactory(ResponseFactory):
    def create_success(self, data, *args, **kwargs):
        inst = super().create_success(data, *args, **kwargs)
        content_data = data.get("content_data")
        descriptions = content_data.get("messages", {}).get("descriptions", [])
        for descr in descriptions:
            code = descr.get("code")
            if code in {416, 413}:
                inst.errors.append((code, descr.get("description")))
        return inst

    def create_fail(self, *args, **kwargs):
        data = args[0]
        error_message = data.get("error_message", None)
        error_code = data.get("error_code", None)
        universes = kwargs.get("universe", None)
        fields = kwargs.get("fields", None)
        if error_code not in response_errors.keys():
            data["error_message"] = response_errors["default"].format(
                error_message=error_message, fields=fields, universes=universes
            )
        else:
            data["error_message"] = response_errors[error_code].format(
                fields=fields, universes=universes
            )

        inst = self.response_class(is_success=False, **data)
        inst.data = self.data_class(data.get("content_data"))
        inst.data._owner = inst
        return inst


class DataGridUDFResponseFactory(ResponseFactory):
    def get_raw(self, data):
        content_data = data.get("content_data", {})
        raw = content_data.get("responses", [])
        raw: Dict = raw[0]
        return raw

    def create_success(self, data, *args, **kwargs):
        inst = super().create_success(data, *args, **kwargs)
        raw = self.get_raw(data)
        error = raw.get("error", [])
        for err in error:
            code = err.get("code")
            if code == 416:
                inst.errors.append((code, err.get("message")))

        return inst

    def create_fail(self, data, *args, **kwargs):
        error_message = data.get("error_message", None)
        error_code = data.get("error_code", None)
        universes = kwargs.get("universe", None)
        fields = kwargs.get("fields", None)

        if error_code not in response_errors.keys():
            data["error_message"] = response_errors["default"].format(
                error_message=error_message, fields=fields, universes=universes
            )

        else:
            data["error_message"] = response_errors[error_code].format(
                fields=fields, universes=universes
            )

        inst = self.response_class(is_success=False, **data)
        inst.data = self.data_class(data.get("content_data"))
        inst.data._owner = inst
        return inst


# --------------------------------------------------------------------------------------
#   Request
# --------------------------------------------------------------------------------------


def validate_correct_format_parameters(*_, **kwargs) -> dict:
    parameters = kwargs.get("parameters")
    extended_params = kwargs.get("extended_params")
    universe = kwargs.get("universe")
    fields = kwargs.get("fields")
    use_field_names_in_headers = kwargs.get("use_field_names_in_headers")

    if parameters is not None and not isinstance(parameters, dict):
        raise ValueError(f"Arg parameters must be a dictionary")

    extended_params = extended_params or {}
    universe = extended_params.get("universe") or universe
    universe = universe_arg_parser.get_list(universe)
    universe = [value.upper() if value.islower() else value for value in universe]
    fields = fields_arg_parser.get_list(fields)
    use_field_names_in_headers = use_field_names_in_headers_arg_parser.get_bool(
        use_field_names_in_headers
    )

    kwargs.update(
        {
            "universe": universe,
            "fields": fields,
            "parameters": parameters,
            "use_field_names_in_headers": use_field_names_in_headers,
            "extended_params": extended_params,
        }
    )
    return kwargs


class DataGridRDPRequestFactory(RequestFactory):
    def get_body_parameters(self, *_, **kwargs) -> dict:
        kwargs = validate_correct_format_parameters(*_, **kwargs)
        body_parameters = {}

        universe = kwargs.get("universe")
        if universe:
            body_parameters["universe"] = universe

        fields = kwargs.get("fields")
        if fields:
            body_parameters["fields"] = fields

        parameters = kwargs.get("parameters")
        if parameters:
            body_parameters["parameters"] = parameters

        layout = kwargs.get("layout")
        if isinstance(layout, dict) and layout.get("output"):
            body_parameters["output"] = layout["output"]

        return body_parameters

    def get_request_method(self, *_, **kwargs) -> RequestMethod:
        return RequestMethod.POST


class DataGridUDFRequestFactory(RequestFactory):
    def create(self, *args, **kwargs):
        session = args[0]
        url_root = session._get_rdp_url_root()
        url = url_root.replace("rdp", "udf")

        method = self.get_request_method(*args, **kwargs)
        header_parameters = kwargs.get("header_parameters") or {}
        extended_params = kwargs.get("extended_params") or {}
        body_parameters = self.get_body_parameters(*args, **kwargs)
        body_parameters = self.extend_body_parameters(body_parameters, extended_params)

        headers = {"Content-Type": "application/json"}
        headers.update(header_parameters)
        request = {
            "url": url,
            "method": request_method_value.get(method),
            "headers": headers,
            "body": {
                "Entity": {
                    "E": "DataGrid_StandardAsync",
                    "W": {"requests": [body_parameters]},
                }
            },
        }

        return request

    def get_body_parameters(self, *_, **kwargs) -> dict:

        ticket = kwargs.get("ticket", None)
        if ticket:
            return {"ticket": ticket}

        kwargs = validate_correct_format_parameters(*_, **kwargs)
        body_parameters = {}

        instruments = kwargs.get("universe")
        if instruments:
            body_parameters["instruments"] = instruments

        fields = kwargs.get("fields")
        if fields:
            body_parameters["fields"] = [
                {"name": i} for i in fields if re.match(ADC_TR_PATTERN, i)
            ]

        parameters = kwargs.get("parameters")
        if parameters:
            body_parameters["parameters"] = parameters

        layout = kwargs.get("layout")
        if isinstance(layout, dict) and layout.get("layout"):
            body_parameters["layout"] = layout["layout"]

        return body_parameters

    def get_request_method(self, *_, **kwargs) -> RequestMethod:
        return RequestMethod.POST


# --------------------------------------------------------------------------------------
#   Content data validator
# --------------------------------------------------------------------------------------


class DataGridRDPContentValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        is_valid = True

        status = data.get("status", {})
        status_content = status.get("content", "")

        if status_content.startswith("Failed"):
            is_valid = False
            data["error_code"] = -1
            data["error_message"] = status_content
            return is_valid

        content_data = data.get("content_data", {})
        content_data_ = content_data.get("data")
        error = content_data.get("error")

        if not content_data:
            is_valid = False
        elif error and not content_data_:
            is_valid = False
            data["error_code"] = error.get("code", -1)
            data["error_message"] = error.get("description")

            if not data["error_message"]:
                error_message = error.get("message")
                errors = error.get("errors")
                if isinstance(errors, list):
                    error_message += ":\n"
                    error_message += "\n".join(map(str, errors))
                data["error_message"] = error_message

        return is_valid


class DataGridUDFContentValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        is_valid = True

        status = data.get("status", {})
        status_content = status.get("content", "")

        if status_content.startswith("Failed"):
            is_valid = False
            data["error_code"] = -1
            data["error_message"] = status_content
            return is_valid

        content_data = data.get("content_data", {})
        if isinstance(content_data, str):
            is_valid = False
            data["error_code"] = -1
            data["error_message"] = content_data
            return is_valid

        response = content_data.get("responses", [])
        response = response[0] if response else {}
        headers = response.get("headers", [])
        headers = headers[0] if headers else {}
        response_data = response.get("data")
        error = response.get("error")

        if not content_data:
            is_valid = False
        elif "ErrorCode" in content_data:
            is_valid = False
            data["error_code"] = content_data.get("ErrorCode", -1)
            data["error_message"] = content_data.get("ErrorMessage")
        elif error and isinstance(error, dict) and not response_data:
            is_valid = False
            data["error_code"] = error.get("code", -1)
            data["error_message"] = error.get("message") or error
        elif error and not response_data:
            is_valid = False
            data["error_code"] = -1
            data["error_message"] = error
        elif headers and error and not all(i.get("displayName") for i in headers):
            is_valid = False
            error_ = error[0]
            data["error_code"] = error_.get("code", -1)
            data["error_message"] = error_.get("message") or error_

        return is_valid


# --------------------------------------------------------------------------------------
#   Providers
# --------------------------------------------------------------------------------------


data_grid_rdp_data_provider = DataProvider(
    request=DataGridRDPRequestFactory(),
    response=DataGridRDPResponseFactory(),
    validator=ValidatorContainer(content_validator=DataGridRDPContentValidator()),
)

data_grid_udf_data_provider = DataProvider(
    request=DataGridUDFRequestFactory(),
    response=DataGridUDFResponseFactory(),
    validator=ValidatorContainer(content_validator=DataGridUDFContentValidator()),
)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.219828

__all__ = ["UpOrDown"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class UpOrDown(Enum):
    DOWN = "Down"
    UP = "Up"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(UpOrDown, _UP_OR_DOWN_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_UP_OR_DOWN_VALUES_IN_LOWER_BY_UP_OR_DOWN, some)


_UP_OR_DOWN_VALUES = tuple(t.value for t in UpOrDown)
_UP_OR_DOWN_VALUES_IN_LOWER_BY_UP_OR_DOWN = {
    name.lower(): item for name, item in UpOrDown.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-18 08:30:59.231843

__all__ = ["PricingModelType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class PricingModelType(Enum):
    BINOMIAL = "Binomial"
    BLACK_SCHOLES = "BlackScholes"
    CEV = "CEV"
    CONTINUOUS = "Continuous"
    LOCAL_VOLATILITY = "LocalVolatility"
    TRINOMIAL = "Trinomial"
    VANNA_VOLGA = "VannaVolga"
    WHALEY = "Whaley"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(PricingModelType, _PRICING_MODEL_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _PRICING_MODEL_TYPE_VALUES_IN_LOWER_BY_PRICING_MODEL_TYPE, some
        )


_PRICING_MODEL_TYPE_VALUES = tuple(t.value for t in PricingModelType)
_PRICING_MODEL_TYPE_VALUES_IN_LOWER_BY_PRICING_MODEL_TYPE = {
    name.lower(): item for name, item in PricingModelType.__members__.items()
}



import re
from collections import Counter
from datetime import datetime, timedelta, date
from typing import Union, Optional, List, Iterable, Callable

import pandas as pd
from numpy import nan
from pandas import DataFrame, Index
from pandas import MultiIndex

from ._tools import get_name_columns_without_datetime
from .get_data import (
    ADC_TR_PATTERN,
    ADC_FUNC_PATTERN_IN_FIELDS,
    _find_and_rename_duplicated_columns,
    _rename_column_n_to_column,
)
from .get_data import _send_request
from .._core.session import get_default
from .._errors import RDError
from .._tools import ohlc, fr_datetime_adapter, DEBUG
from .._tools._dataframe import convert_df_columns_to_datetime
from ..content import fundamental_and_reference, custom_instruments
from ..content import historical_pricing
from ..content.historical_pricing._hp_data_provider import EventTypes

EVENTS_INTERVALS = ["tick", "tas", "taq"]

PRICING_PARAMS_NAMES = [
    "universe",
    "fields",
    "interval",
    "start",
    "end",
    "adjustments",
    "count",
    "eventTypes",
]
CUSTOM_INSTS_PARAMS_NAMES = ["universe", "start", "end", "count", "interval"]

INTERVALS = {
    "tick": {"event_types": None, "adc": "D"},
    "tas": {"event_types": EventTypes.TRADE, "adc": "D"},
    "taq": {"event_types": EventTypes.QUOTE, "adc": "D"},
    "minute": {"pricing": "PT1M", "adc": "D"},
    "1min": {"pricing": "PT1M", "adc": "D"},
    "5min": {"pricing": "PT5M", "adc": "D"},
    "10min": {"pricing": "PT10M", "adc": "D"},
    "30min": {"pricing": "PT30M", "adc": "D"},
    "60min": {"pricing": "PT60M", "adc": "D"},
    "hourly": {"pricing": "PT1H", "adc": "D"},
    "1h": {"pricing": "PT1H", "adc": "D"},
    "daily": {"pricing": "P1D", "adc": "D"},
    "1d": {"pricing": "P1D", "adc": "D"},
    "1D": {"pricing": "P1D", "adc": "D"},
    "7D": {"pricing": "P7D", "adc": "W"},
    "7d": {"pricing": "P7D", "adc": "W"},
    "weekly": {"pricing": "P1W", "adc": "W"},
    "1W": {"pricing": "P1W", "adc": "W"},
    "monthly": {"pricing": "P1M", "adc": "M"},
    "1M": {"pricing": "P1M", "adc": "M"},
    "quarterly": {"pricing": "P3M", "adc": "CQ"},
    "3M": {"pricing": "P3M", "adc": "CQ"},
    "6M": {"pricing": "P6M", "adc": "CS"},
    "yearly": {"pricing": "P1Y", "adc": "CY"},
    "1Y": {"pricing": "P1Y", "adc": "CY"},
}

NON_INTRA_DAY_INTERVALS = {
    "daily",
    "1d",
    "1D",
    "weekly",
    "7D",
    "7d",
    "1W",
    "monthly",
    "1M",
    "quarterly",
    "3M",
    "yearly",
    "1Y",
}


def get_history(
    universe: Union[str, list],
    fields: Union[str, list, None] = None,
    interval: Optional[str] = None,
    start: Union[str, date, datetime, timedelta] = None,
    end: Union[str, date, datetime, timedelta] = None,
    adjustments: Optional[str] = None,
    count: Optional[int] = None,
    use_field_names_in_headers: Optional[bool] = False,
) -> DataFrame:
    """
    With this tool you can request historical data from Pricing and ADC

    Parameters
    ----------
        universe: str | list
            instruments to request.
        fields: str | list, optional
            fields to request.
        interval: str, optional
            The consolidation interval. Supported intervals are:
            tick, tas, taq, minute, 1min, 5min, 10min, 30min, 60min, hourly, 1h, daily,
            1d, 1D, 7D, 7d, weekly, 1W, monthly, 1M, quarterly, 3M, 6M, yearly, 1Y
        start: str or date or datetime or timedelta, optional
            The start date and timestamp of the query in ISO8601 with UTC only
        end: str or date or datetime or timedelta, optional
            The end date and timestamp of the query in ISO8601 with UTC only
        adjustments : str, optional
            The adjustment
        count : int, optional
            The maximum number of data returned. Values range: 1 - 10000
        use_field_names_in_headers : bool, optional
            Return field name in headers instead of title

    Returns
    -------
    pandas.DataFrame

     Examples
    --------
    >>> get_history(universe="GOOG.O")
    >>> get_history(universe="GOOG.O", fields="tr.Revenue", interval="1Y")
    >>> get_history(
    ...     universe="GOOG.O",
    ...     fields=["BID", "ASK", "tr.Revenue"],
    ...     interval="1Y",
    ...     start="2015-01-01",
    ...     end="2020-10-01",
    ... )
    """
    if interval not in INTERVALS and interval is not None:
        raise ValueError(
            f"Not supported interval value.\nSupported intervals are:"
            f"{list(INTERVALS.keys())}"
        )

    _pricing_events = historical_pricing.events.Definition
    _pricing_summaries = historical_pricing.summaries.Definition
    _ci_events = custom_instruments.events.Definition
    _ci_summaries = custom_instruments.summaries.Definition
    _fundamental_data = fundamental_and_reference.Definition
    if isinstance(universe, str):
        universe = [universe]
    universe = list(dict.fromkeys(universe))
    if not fields:
        fields = []
    if isinstance(fields, str):
        fields = [fields]
    params = {
        "universe": universe,
        "fields": fields,
        "interval": interval,
        "start": start,
        "end": end,
        "adjustments": adjustments,
        "count": count,
        "use_field_names_in_headers": use_field_names_in_headers,
    }

    if interval in EVENTS_INTERVALS:
        p_provider = _pricing_events
        ci_provider = _ci_events
        params.pop("interval")
        params["eventTypes"] = INTERVALS[interval]["event_types"]
        index_name = "Timestamp"

    else:
        p_provider = _pricing_summaries
        ci_provider = _ci_summaries
        index_name = "Date"
    return _get_history(_fundamental_data, p_provider, ci_provider, params, index_name)


def _get_history(
    adc_provider: Callable,
    p_provider: Callable,
    ci_provider: Callable,
    params: dict,
    index_name: str,
) -> DataFrame:
    logger = get_default().logger()
    fields = params.get("fields")
    use_field_names_in_headers = params.get("use_field_names_in_headers")
    universe = params.get("universe")
    interval = params.get("interval")

    dfs = []
    duplicated_columns = []

    adc_params = get_adc_params(params)
    adc_and_pricing_universe = adc_params.get("universe")
    adc_fields = adc_params.get("fields")
    not_adc_fields = [i for i in fields if i not in adc_fields]
    universe_for_template = []

    if not adc_fields:
        adc_params["fields"] = ["TR.RIC"]

    # adc
    if adc_and_pricing_universe:
        _adc_df = _send_request(
            data_provider=adc_provider,
            params=adc_params,
            logger=logger,
        )
        DEBUG and logger.debug(f"DATAGRID --->\n{_adc_df.to_string()}\n")
        # update universe
        adc_and_pricing_universe = (
            get_universe_from_df(_adc_df) or adc_and_pricing_universe
        )
        universe_for_template.extend(adc_and_pricing_universe)
        if adc_fields:
            adc_df = _adc_df
            duplicated_columns = _find_and_rename_duplicated_columns(adc_df)
            if duplicated_columns:
                remove_duplicated_columns(adc_df, duplicated_columns)

            adc_df = to_multi_index(adc_df, adc_and_pricing_universe)
            dfs.append(adc_df)

        # pricing
        pricing_params = get_pricing_params(
            params, adc_and_pricing_universe, adc_fields
        )
        pricing_fields = pricing_params.get("fields")
        if not fields or pricing_fields:
            pricing_df = _send_request(
                data_provider=p_provider, params=pricing_params, logger=logger
            )
            pricing_df = to_multi_index(pricing_df, adc_and_pricing_universe)
            DEBUG and logger.debug(f"PRICING --->\n{pricing_df.to_string()}\n")
            if fields:
                _remove_fields_if_not_requested(
                    fields, use_field_names_in_headers, pricing_df
                )
            dfs.append(pricing_df)

    # custom_instruments
    ci_params = get_ci_params(params)
    ci_universe = ci_params.get("universe")
    if ci_universe:
        universe_for_template.extend(ci_universe)

        ci_df = _send_request(
            data_provider=ci_provider,
            params=ci_params,
            logger=logger,
        )
        DEBUG and logger.debug(f"CUSTOMINSTS --->\n{ci_df.to_string()}\n")
        ci_df = to_multi_index(ci_df, ci_universe)
        if fields:
            _remove_fields_if_not_requested(fields, use_field_names_in_headers, ci_df)

        dfs.append(ci_df)

    # join dfs
    _look_for_exceptions(dfs)
    universe_for_template.sort(
        key=lambda a: universe.index(a) if a in universe else float("inf")
    )
    df_template = get_df_template(
        universe_for_template, not_adc_fields, use_field_names_in_headers
    )

    result = join_dfs(dfs, df_template, universe, interval, index_name)
    for i in duplicated_columns:
        _rename_column_n_to_column(i, result, multiindex=True)

    result = _update_columns_title(result)
    result.ohlc = ohlc.__get__(result, None)

    return result


def remove_duplicated_columns(df, duplicated_columns):
    if df.index.name in {"Date", "date"}:
        date_column = df.index.name
    else:
        date_column = "Date"

    if date_column in duplicated_columns:
        date_column = f"{date_column}_0"

    convert_df_columns_to_datetime(df, pattern=date_column, utc=True, delete_tz=True)
    df.drop_duplicates(inplace=True)
    return df


def _remove_fields_if_not_requested(fields, use_field_names_in_headers, df):
    fields_list = fields
    if not use_field_names_in_headers:
        fields_list = [i[3:] if i.upper().startswith("TR.") else i for i in fields]

    fields_from_df = {i[1] for i in df.columns}
    for field in fields_from_df:
        field_not_in_fields_list = str(field).upper() not in {
            i.upper() for i in fields_list
        }
        if not df.empty and field_not_in_fields_list:
            df.drop(field, axis=1, level=1, inplace=True, errors="ignore")


def get_df_template(universe, fields, use_field_names_in_headers):
    if not universe or not fields:
        return DataFrame()

    if use_field_names_in_headers:
        fields_for_template = fields
    else:
        fields_for_template = [
            i[3:] if i.upper().startswith("TR.") else i for i in fields
        ]

    columns = pd.MultiIndex.from_tuples(
        [(name, field) for field in fields_for_template for name in universe]
    )
    df_template = DataFrame(columns=columns)
    return df_template


def join_dfs(dfs, df_template, universe, interval, index_name):
    result = dfs.pop(0)
    for df in dfs:
        result = result.join(df, how="outer")

    columns = get_name_columns_without_datetime(result)
    if not result.empty:
        result[columns] = result[columns].replace([pd.NaT, nan], pd.NA)
    result = pd.concat([df_template, result])

    result.sort_index(ascending=True, inplace=True)

    if isinstance(result.columns, MultiIndex):
        result = result.reindex(columns=sorted(result.columns))

    result = _sort_by_universe(result, universe)
    result.index.name = index_name
    result = _set_index_name_and_change_index_type(result, interval)
    return result


def filter_params_by_name(params: dict, names: Iterable[str]):
    result = {key: value for key, value in params.items() if key in names and value}
    return result


def get_ci_params(params):
    universe = params.get("universe", [])
    universe = [i for i in universe if i.startswith("S)")]
    custom_insts_params = filter_params_by_name(params, CUSTOM_INSTS_PARAMS_NAMES)
    custom_insts_params["universe"] = universe
    interval = custom_insts_params.get("interval")
    if interval is not None:
        custom_insts_params["interval"] = INTERVALS[custom_insts_params["interval"]][
            "pricing"
        ]
    return custom_insts_params


def get_pricing_params(params, universe, adc_fields):
    fields = params.get("fields")
    if fields:
        fields = [i for i in fields if i not in adc_fields]
    pricing_params = filter_params_by_name(params, PRICING_PARAMS_NAMES)
    interval = pricing_params.get("interval")
    if interval is not None:
        pricing_params["interval"] = INTERVALS[pricing_params["interval"]]["pricing"]
    pricing_params["universe"] = universe
    pricing_params["fields"] = fields
    return pricing_params


def get_adc_params(params: dict):
    fields = params.get("fields")
    _adc_params = _translate_pricing_params_to_adc(params)
    adc_fields = get_adc_fields(fields) if fields else []
    universe = params.get("universe", [])
    universe = [i for i in universe if not i.startswith("S)")]
    adc_params = {
        "universe": universe,
        "fields": adc_fields,
        "parameters": _adc_params,
        "row_headers": "date",
        "use_field_names_in_headers": params.get("use_field_names_in_headers"),
    }
    return adc_params


def get_universe_from_df(df):
    if "Instrument" in df:
        instrument = df["Instrument"]
        result = list(instrument)
    elif "RIC" in df:
        instrument = df["RIC"]
        result = list(instrument)
    else:
        columns = [column[0] for column in df.columns if isinstance(column, tuple)]
        result = list(dict.fromkeys(columns))
    return result


def get_adc_fields(fields):
    adc_tr_fields = [i for i in fields if re.match(ADC_TR_PATTERN, i)]
    adc_funcs_in_fields = [i for i in fields if re.match(ADC_FUNC_PATTERN_IN_FIELDS, i)]

    adc_fields = adc_tr_fields + adc_funcs_in_fields
    return adc_fields


def _look_for_exceptions(dfs: list):
    exceptions = []
    raise_exception = True
    for df in dfs:
        if (
            hasattr(df.flags, "exception_event")
            and df.flags.exception_event["raise_exception"]
        ):
            _exceptions = [
                i.message
                for i in df.flags.exception_event["exception"]
                if hasattr(i, "message")
            ]
            exceptions.extend(_exceptions)
        else:
            raise_exception = False
    if raise_exception:
        exceptions = "\n".join(exceptions)
        raise RDError(1, f"\nNo data to return, please check errors:\n {exceptions}")


def to_multi_index(df: DataFrame, universe: List[str]):
    _df = df
    if not isinstance(_df.columns, pd.MultiIndex):
        if len(universe) == 1:
            _df.columns = pd.MultiIndex.from_product([[universe[0]], _df.columns])
        elif _df.columns.name is not None:
            field = _df.columns.name
            _df.columns = pd.MultiIndex.from_tuples(
                [(_universe, field) for _universe in _df]
            )
        _df.columns.names = [None] * len(df.columns.names)
    return df


def _sort_by_universe(df: DataFrame, universe: List[str]) -> DataFrame:
    length = len(universe)

    if length == 1:
        return df

    columns = df.columns

    def make_getidx():
        get_index = universe.index
        if isinstance(columns, MultiIndex):

            def geti(i):
                return i[0]

        else:

            def geti(i):
                return i

        def inner(i):
            try:
                index = get_index(geti(i))
            except ValueError:
                index = length
            return index

        return inner

    getidx = make_getidx()
    # [3, 0, 2, 1]
    curr_order = [getidx(col) for col in columns]
    # [0, 1, 2, 3]
    expected_order = list(range(length))
    if curr_order != expected_order:
        sorted_columns = (col for _, col in sorted(zip(curr_order, columns)))
        df = df.reindex(columns=sorted_columns)
    return df


def _update_columns_title(df: DataFrame) -> DataFrame:
    """
    Transform dataframe from:

                AAPL.O          IBM.N
                Revenue - Mean  Revenue - Mean
    Date
    2022-04-12  396147651250    <NA>
    2022-04-18  <NA>            60686514370

    to:

    Revenue - Mean  AAPL.O            IBM.N
    Date
    2022-04-12      396147651250      <NA>
    2022-04-18      <NA>              60686514370
    """

    columns = df.columns
    if isinstance(columns, MultiIndex):

        c_fields = Counter(col[1] for col in columns if len(col) == 2)
        c_universe = Counter(col[0] for col in columns if len(col) == 2)
        fields_keys = list(c_fields.keys())
        universe_keys = list(c_universe.keys())
        if len(universe_keys) == 1:
            df.columns = Index([col[1] for col in columns], name=universe_keys[0])
        elif len(fields_keys) == 1:
            df.columns = Index([col[0] for col in columns], name=fields_keys[0])

    return df


def _translate_pricing_params_to_adc(p_params: dict) -> dict:
    adc_params = {}

    start = p_params.get("start")
    if start is not None:
        adc_params["SDate"] = fr_datetime_adapter.get_str(start)

    end = p_params.get("end")
    if end is not None:
        adc_params["EDate"] = fr_datetime_adapter.get_str(end)

    interval = p_params.get("interval")
    if interval is not None:
        adc_params["FRQ"] = INTERVALS[interval]["adc"]

    return adc_params


def _set_index_name_and_change_index_type(df: DataFrame, interval):
    if interval is not None and interval not in NON_INTRA_DAY_INTERVALS:
        df.index.names = ["Timestamp"]

    if interval in NON_INTRA_DAY_INTERVALS or interval is None:

        if isinstance(df.index, pd.Index):
            df.index = pd.to_datetime(df.index)

        df.index = df.index.tz_localize(None)

    return df



# coding: utf8
# contract_gen 2020-05-18 08:30:59.213816

__all__ = ["YieldType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class YieldType(Enum):
    """
    YieldType that specifies the rate structure:
    - Native : no specific yield type is defined.
    - UsGovt_Actual_Actual_6M : US Govt Act/Act 6M YTA.
    - Isma_30_360_6M : ISMA 30/360 6M YTA.
    - Euroland_Actual_Actual_6M : Euroland Equivalent Act/Act 6M YTA.
    - Money_Market_Actual_360_6M : Money Market Act/360 6M YTA.
    - Money_Market_Actual_365_6M : Money Market Act/365 6M YTA.
    - Money_Market_Actual_Actual_6M : Money Market Act/Act 6M YTA.
    - Bond_Actual_364_6M : Bond Market Act/364 6M YTA.
    - Japanese_Simple_JAP_6M : Japanese Simple JAP 6M YTA.
    - Japanese_Compunded_30_360_6M : Japanese Compounded 30/360 6M YTA.
    - Moosmueller_30_360_6M : Moosmueller 30/360 6M YTA.
    - Braess_Frangmeyer_30_360_6M : Braess-Frangmeyer 30/360 6M YTA.
    - Weekend_30_360 : Week End 30/360 6M YTA.
    """

    ANNUAL_EQUIVALENT = "Annual_Equivalent"
    BOND_ACTUAL_364 = "Bond_Actual_364"
    BRAESS_FANGMEYER = "Braess_Fangmeyer"
    DISCOUNT_ACTUAL_360 = "Discount_Actual_360"
    DISCOUNT_ACTUAL_365 = "Discount_Actual_365"
    EUROLAND = "Euroland"
    ISMA = "Isma"
    JAPANESE_COMPOUNDED = "Japanese_Compounded"
    JAPANESE_SIMPLE = "Japanese_Simple"
    MONEY_MARKET_ACTUAL_360 = "MoneyMarket_Actual_360"
    MONEY_MARKET_ACTUAL_365 = "MoneyMarket_Actual_365"
    MONEY_MARKET_ACTUAL_ACTUAL = "MoneyMarket_Actual_Actual"
    MOOSMUELLER = "Moosmueller"
    NATIVE = "Native"
    QUARTERLY_EQUIVALENT = "Quarterly_Equivalent"
    SEMIANNUAL_EQUIVALENT = "Semiannual_Equivalent"
    TURKISH_COMPOUNDED = "TurkishCompounded"
    US_GOVT = "UsGovt"
    US_T_BILLS = "UsTBills"
    WEEKEND = "Weekend"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(YieldType, _YIELD_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_YIELD_TYPE_VALUES_IN_LOWER_BY_YIELD_TYPE, some)


_YIELD_TYPE_VALUES = tuple(t.value for t in YieldType)
_YIELD_TYPE_VALUES_IN_LOWER_BY_YIELD_TYPE = {
    name.lower(): item for name, item in YieldType.__members__.items()
}



# coding: utf8

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class AmortizationType(Enum):
    ANNUITY = "Annuity"
    LINEAR = "Linear"
    NONE = "None"
    SCHEDULE = "Schedule"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(AmortizationType, _AMORTIZATION_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_AMORTIZATION_TYPE_VALUES_IN_LOWER_BY_AMORTIZATION_TYPE, some)


_AMORTIZATION_TYPE_VALUES = tuple(t.value for t in AmortizationType)
_AMORTIZATION_TYPE_VALUES_IN_LOWER_BY_AMORTIZATION_TYPE = {
    name.lower(): item for name, item in AmortizationType.__members__.items()
}



# coding: utf8

from enum import Enum, unique

from ..._enums._common_tools import _convert_to_str, _normalize


@unique
class ZcInterpolationMode(Enum):
    AKIMA_METHOD = "AkimaMethod"
    CUBIC_DISCOUNT = "CubicDiscount"
    CUBIC_RATE = "CubicRate"
    CUBIC_SPLINE = "CubicSpline"
    FORWARD_MONOTONE_CONVEX = "ForwardMonotoneConvex"
    FRITSCH_BUTLAND_METHOD = "FritschButlandMethod"
    HERMITE = "Hermite"
    KRUGER_METHOD = "KrugerMethod"
    LINEAR = "Linear"
    LOG = "Log"
    MONOTONIC_CUBIC_NATURAL_SPLINE = "MonotonicCubicNaturalSpline"
    MONOTONIC_HERMITE_CUBIC = "MonotonicHermiteCubic"
    TENSION_SPLINE = "TensionSpline"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(ZcInterpolationMode, _ZC_INTERPOLATION_MODE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _ZC_INTERPOLATION_MODE_VALUES_IN_LOWER_BY_ZC_INTERPOLATION_MODE, some
        )


_ZC_INTERPOLATION_MODE_VALUES = tuple(t.value for t in ZcInterpolationMode)
_ZC_INTERPOLATION_MODE_VALUES_IN_LOWER_BY_ZC_INTERPOLATION_MODE = {
    name.lower(): item for name, item in ZcInterpolationMode.__members__.items()
}



# coding: utf8
# contract_gen 2020-05-18 08:30:59.230815

__all__ = ["TimeStamp"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class TimeStamp(Enum):
    DEFAULT = "Default"
    OPEN = "Open"
    CLOSE = "Close"
    SETTLE = "Settle"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(TimeStamp, _TIME_STAMP_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_TIME_STAMP_VALUES_IN_LOWER_BY_TIME_STAMP, some)


_TIME_STAMP_VALUES = tuple(t.value for t in TimeStamp)
_TIME_STAMP_VALUES_IN_LOWER_BY_TIME_STAMP = {
    name.lower(): item for name, item in TimeStamp.__members__.items()
}



from types import SimpleNamespace
from typing import List, Callable, Tuple

import numpy as np
import pandas as pd
from pandas import DataFrame

from ..delivery._data._data_provider import Response, Data


def join_dfs(dfs: List[DataFrame], how: str = "inner") -> DataFrame:
    if len(dfs) == 0:
        raise ValueError(f"Cannot join dfs, because dfs list is empty, dfs={dfs}")

    df = dfs.pop()
    df = df.join(dfs, how=how)  # noqa
    df = df.convert_dtypes()

    return df


def join_responses(
    responses: List[Response],
    join_dataframes: Callable = pd.concat,
    response_class=Response,
    data_class=Data,
    reset_index=False,
) -> Response:
    def build_df(*args, **kwargs):
        dfs = []
        df = None

        for response in responses:
            dfs.append(response.data.df)

        all_dfs_is_none = all(a is None for a in dfs)
        if not all_dfs_is_none:
            df = join_dataframes(dfs)

        if reset_index and df is not None:
            df = df.reset_index(drop=True)

        return df

    if len(responses) == 1:
        return responses[0]

    raws = []
    http_statuses = []
    http_headers = []
    request_messages = []
    http_responses = []
    errors = []
    is_successes = []

    for response in responses:
        raws.append(response.data.raw)
        http_statuses.append(response.http_status)
        http_headers.append(response.http_headers)
        request_messages.append(response.request_message)
        http_responses.append(response.http_response)
        is_successes.append(response.is_success)

        if response.errors:
            errors += response.errors

    raw_response = SimpleNamespace()
    raw_response.headers = http_headers
    raw_response.request = request_messages
    is_success = any(is_successes)
    response = response_class(raw_response=raw_response, is_success=is_success)
    response.data = data_class(raws, dfbuilder=build_df)
    response.errors += errors
    response.http_response = http_responses
    response._status = http_statuses

    return response


def get_first_success_response(responses: List[Tuple[str, Response]]) -> Response:
    successful = (response for _, response in responses if response.is_success)
    first_successful = next(successful, None)
    return first_successful


def join_historical_responses(
    responses: List[Tuple[str, Response]],
    axis_name: str,
    fields: List[str],
) -> Response:
    if len(responses) == 1:
        inst_name, response = responses[0]

        if not response.is_success:
            return response

        def build_df(*args, **kwargs) -> DataFrame:
            df = data.df

            if fields:
                not_valid_columns = set(fields) - set(df.columns.values)
                df = df.assign(
                    **{column_name: pd.NA for column_name in not_valid_columns}
                )

            df.axes[1].name = inst_name
            df.index.name = axis_name
            return df

        data = response.data
        response.data = Data(response.data.raw, dfbuilder=build_df)
        return response

    def build_df_as_join_dfs(*args, **kwargs) -> DataFrame:
        response = get_first_success_response(responses)

        if not response:
            return DataFrame()

        df = response.data.df

        index = df.index.to_numpy()
        columns = (None,)

        dfs = []
        for inst_name, response in responses:
            df = response.data.df

            if fields and response.is_success:
                not_valid_columns = set(fields) - set(df.columns.values)
                df = df.assign(
                    **{column_name: np.NaN for column_name in not_valid_columns}
                )

            elif fields and df is None:
                df = DataFrame(columns=fields, index=index)

            elif df is None:
                df = DataFrame(columns=columns, index=index)

            df.columns = pd.MultiIndex.from_product([[inst_name], df.columns])
            dfs.append(df)

        df = join_dfs(dfs, how="outer")

        # when only one field
        if fields:
            _fields = fields
        else:
            _fields = list(dict.fromkeys([i[1] for i in df.columns]))
        if len(_fields) == 1:
            df.columns = [column[0] for column in df.columns]
            df.columns.name = _fields[0]

        if not df.empty:
            df = df.rename_axis(axis_name)
        return df

    raws = []
    errors = []
    http_statuses = []
    http_headers = []
    http_responses = []
    request_messages = []

    for inst_name, response in responses:
        raws.append(response.data.raw)
        http_statuses.append(response.http_status)
        http_headers.append(response.http_headers)
        request_messages.append(response.request_message)
        http_responses.append(response.http_response)

        if response.errors:
            errors += response.errors

    raw_response = SimpleNamespace()
    raw_response.request = request_messages
    raw_response.headers = http_headers
    response = Response(raw_response=raw_response, is_success=True)
    response.errors += errors
    response.data = Data(raws, dfbuilder=build_df_as_join_dfs)
    response._status = http_statuses
    response.http_response = http_responses

    return response



# coding: utf-8

import collections


class TradeDataStreamCache(object):
    """cache for trading data stream including order summary and order event"""

    #   default cache size
    #       order summary
    _default_order_summary_cache_size = 1000
    #       order events
    _default_order_events_cache_size = 1000

    def __init__(
        self, session: object, max_order_summary_cache: int, max_order_events_cache: int
    ):
        self._session = session

        #   cache size
        #       order summary
        self._max_order_summary_cache = (
            max_order_summary_cache
            if max_order_summary_cache is not None
            else self._default_order_summary_cache_size
        )
        #       order events
        self._max_order_events_cache = (
            max_order_events_cache
            if max_order_events_cache is not None
            else self._default_order_events_cache_size
        )

        #   header
        self._order_summary_header_names = None

        #   cache
        #       order summary
        self._order_summary_dict = collections.OrderedDict()
        #       order events
        self._order_key_to_event_list = {}

    def _clear_caches(self):
        """clear the order summary and order events caches"""
        self._order_summary_dict = collections.OrderedDict()
        self._order_key_to_event_list = {}

    def _add_order_summary(self, key, order_summary: list):
        """add or update the order summary"""

        #   check for cache enable or not?
        if self._max_order_summary_cache == 0:
            #   skip
            return

        #   check for update the old order summary
        if key in self._order_summary_dict:
            #   remove old order summary before add the new one
            #       this for the last update order summary in the dict
            del self._order_summary_dict[key]

        #   check the cache size
        if len(self._order_summary_dict) + 1 > self._max_order_summary_cache:
            #   the data exceeds cache size, so remove oldest data in cache
            self._order_summary_dict.popitem(last=False)

        #   add or update the order summary
        self._order_summary_dict[key] = order_summary

        assert len(self._order_summary_dict) <= self._max_order_summary_cache

    def _update_order_summary(self, order_key, order_summary: dict):
        assert "key" in order_summary
        assert "OrderKey" not in order_summary

        #   check for cache enable or not?
        if self._max_order_summary_cache == 0:
            #   skip
            return

        #   get the 'OrderKey'. it use key as a name
        order_summary_order_key = order_summary.pop("key")

        #   update order summary
        for key, value in order_summary.items():
            try:
                index = self._order_summary_header_names.index(key)
            except ValueError:
                self._session.debug(
                    f"#WARNING!!! received no subscribed field name [{key}]"
                )
                continue

            self._order_summary_dict[order_summary_order_key][index] = value

    def _add_order_event(self, order_event: list):
        assert "key" in order_event
        assert "events" in order_event

        #   check for cache enable or not?
        if self._max_order_events_cache == 0:
            #   skip
            return

        #   extract key and event
        #       key
        key = order_event.get("key", None)
        assert key is not None
        assert isinstance(key, str)

        #       events
        events = order_event.get("events", None)
        assert events is not None
        assert isinstance(events, list)

        #   store in cache
        this_order_events = self._order_key_to_event_list.setdefault(
            key, collections.deque(maxlen=self._max_order_events_cache)
        )
        for event in events:
            #   append to the order event list for this key
            this_order_events.append(event)

        #   done

    def get_last_order_events(self):
        """get the last order events"""

        #   do extract all order events
        all_order_events = []
        for order_key in self._order_key_to_event_list.keys():

            this_order_events = self.__get_last_order_event_with_column_name(order_key)
            for this_order_event in this_order_events:
                assert len(this_order_event) == len(self._event_column_names)

                this_order_event_with_order_key = [
                    order_key,
                ] + this_order_event
                all_order_events.append(this_order_event_with_order_key)

        #   done
        return all_order_events

    @property
    def __order_event_data_column_names(self):
        """the different event type has different data object structure:
        Executed:
        {
            "ExecutionId" : "...",
            "ExecutionQuantity" : "...",
            "ExecutionPrice" : "...",
            "ExecutionValue" : "...",
            "LastMarketMIC": "..."
        }

        Accepted / Replaced / UpdatePending
        {
            "OrderQuantity" : "...",
            "LimitPrice" : "...",
            "StopPrice" : "...",
            "DestinationType": "...",
            "AlgoId": "...",
            "TimeInForce": "...",
            "LastCapacity": "..."
        }

        Cancelled / Rejected / CancelPending

        {
            "Reason" : "..."
        }
        """
        return [
            "ExecutionId",
            "ExecutionQuantity",
            "ExecutionPrice",
            "ExecutionValue",
            "LastMarketMIC",  # Executed
            "OrderQuantity",
            "LimitPrice",
            "StopPrice",
            "DestinationType",
            "AlgoId",
            "TimeInForce",
            "LastCapacity",  # Accepted / Replaced / UpdatePending
            "Reason",  # Cancelled / Rejected / CancelPending
        ]

    @property
    def _event_column_names(self):
        event_column_names = [
            "EventTime",
            "OrderId",
            "EventType",
        ]
        event_column_names.extend(self.__order_event_data_column_names)
        return event_column_names

    def __get_last_order_events(self, order_key: str):
        assert order_key in self._order_key_to_event_list
        return self._order_key_to_event_list[order_key]

    def __get_last_order_event_with_column_name(self, order_key: str):
        """return a of the last order event data from given order key.
        note that the order of data data will be matched with event data column name.
        """

        #   extract and return event data
        events = self.__get_last_order_events(order_key)

        #   loop over all event of this order key
        last_order_event_list = []
        for event in events:
            #   extract last event information
            assert "EventTime" in event
            event_time = event["EventTime"]
            assert "OrderId" in event
            order_id = event["OrderId"]
            assert "EventType" in event
            event_type = event["EventType"]
            assert "EventData" in event
            event_data = event["EventData"]

            #   this order event column name
            this_order_event_column_names = [
                event_data.get(column_name, None)
                for column_name in self.__order_event_data_column_names
            ]

            #   reformat the the event data with column name
            this_event_list = [
                event_time,
                order_id,
                event_type,
            ] + this_order_event_column_names
            last_order_event_list.append(this_event_list)

        #   done
        return last_order_event_list



# coding: utf8

__all__ = ["FxCrossType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class FxCrossType(Enum):
    """
    The type of the Fx Cross instrument :
        'FxSpot',
        'FxForward',
        'FxNonDeliverableForward',
        'FxSwap',
        'MultiLeg' or 'FxForwardForward'.
    """

    FX_FORWARD = "FxForward"
    FX_FWD_FWD = "FxFwdFwd"
    FX_NON_DELIVERABLE_FORWARD = "FxNonDeliverableForward"
    FX_SPOT = "FxSpot"
    FX_SWAP = "FxSwap"
    FX_TIME_OPTION_FORWARD = "FxTimeOptionForward"
    MULTI_LEG = "MultiLeg"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(FxCrossType, _FX_CROSS_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_FX_CROSS_TYPE_VALUES_IN_LOWER_BY_FX_CROSS_TYPE, some)


_FX_CROSS_TYPE_VALUES = tuple(t.value for t in FxCrossType)
_FX_CROSS_TYPE_VALUES_IN_LOWER_BY_FX_CROSS_TYPE = {
    name.lower(): item for name, item in FxCrossType.__members__.items()
}



import os
from typing import Optional
import warnings
import re

import ema
from ema import LoginReq, OmmArray, ReqMsg, Map, ElementList, MapEntry


class IntArray(OmmArray):
    def __init__(self, values, fixed_width=2):
        super().__init__()
        self.fixed_width(fixed_width)
        for value in values:
            self.add_int(value)


# Regular expression pattern for matching ip address
IP_PATTERN = re.compile(r"^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)(\.(?!$)|$)){4}$")


def generate_login_msg(login_msg: dict) -> dict:
    login_msg_kwargs = dict(
        application_id=login_msg["Key"]["Elements"]["ApplicationId"],
        position=login_msg["Key"]["Elements"]["Position"],
    )
    if login_msg["Key"].get("NameType"):
        login_msg_kwargs["name_type"] = name_type_map.get(login_msg["Key"]["NameType"])
        if login_msg_kwargs["name_type"] == ema.USER_AUTH_TOKEN:
            login_msg_kwargs["name"] = login_msg["Key"]["Elements"][
                "AuthenticationToken"
            ]
    if "name" not in login_msg_kwargs:
        login_msg_kwargs["name"] = login_msg["Key"]["Name"]
    return login_msg_kwargs


def ema_login_message(
    application_id, position, name, name_type=ema.USER_NAME
) -> ReqMsg:
    return (
        LoginReq()
        .name(name)
        .name_type(name_type)
        .position(position)
        .application_id(application_id)
        .get_message()
    )


severity_map = {
    "verbose": ema.LoggerSeverity.Verbose.value,
    "success": ema.LoggerSeverity.Success.value,
    "warning": ema.LoggerSeverity.Warning.value,
    "error": ema.LoggerSeverity.Error.value,
    "silent": ema.LoggerSeverity.NoLogMsg.value,
}

# TODO: Find out other possible name types in json
name_type_map = {
    "AuthnToken": ema.USER_AUTH_TOKEN,
}


def logger_severity_from_env():
    logger_severity_string = os.getenv("EMA_LOGGER_SEVERITY", "silent").lower()
    if logger_severity_string not in severity_map:
        return ema.LoggerSeverity.NoLogMsg.value
    else:
        return severity_map[logger_severity_string]


def create_programmatic_cfg(
    *,
    field_dict_path: Optional[str] = None,
    enumtype_path: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
):
    config_map = Map()
    inner_map = Map()
    element_list = ElementList()

    element_list.add_ascii("DefaultConsumer", "Cons_Token")
    inner_map.add_key_ascii(
        "Cons_Token",
        MapEntry.AddEnum,
        ElementList()
        .add_ascii("Channel", "Chan_Token")
        .add_ascii("Logger", "Logger_1")
        .add_ascii("Dictionary", "Dictionary_1")
        .add_uint("XmlTraceToStdout", 0)
        .complete(),
    ).complete()

    element_list.add_map("ConsumerList", inner_map)

    element_list.complete()
    inner_map.clear()

    config_map.add_key_ascii("ConsumerGroup", MapEntry.AddEnum, element_list)
    element_list.clear()

    if IP_PATTERN.match(host):
        channel_type = ema.RSSL_CONN_TYPE_SOCKET
    else:
        channel_type = ema.RSSL_CONN_TYPE_ENCRYPTED
    channel_element_list = (
        ElementList()
        .add_enum("ChannelType", channel_type.value)
        .add_enum("CompressionType", ema.RSSL_COMP_NONE.value)
        .add_uint("ConnectionPingTimeout", 1000)
        .add_uint("EnableSessionManagement", 0)
        # .add_uint("GuaranteedOutputBuffers", 5000)
        .add_uint("EncryptedProtocolType", ema.RSSL_CONN_TYPE_SOCKET.value)
    )

    if host is not None:
        channel_element_list.add_ascii("Host", host)
    if port is not None:
        channel_element_list.add_uint("Port", port)
    inner_map.add_key_ascii(
        "Chan_Token", MapEntry.AddEnum, channel_element_list.complete()
    ).complete()

    element_list.add_map("ChannelList", inner_map)

    element_list.complete()
    inner_map.clear()

    config_map.add_key_ascii("ChannelGroup", ema.MapEntry.AddEnum, element_list)
    element_list.clear()

    inner_map.add_key_ascii(
        "Logger_1",
        ema.MapEntry.AddEnum,
        ema.ElementList()
        .add_enum("LoggerType", ema.LoggerType.Stdout.value)
        .add_enum("LoggerSeverity", logger_severity_from_env())
        .complete(),
    ).complete()

    element_list.add_map("LoggerList", inner_map)

    element_list.complete()
    inner_map.clear()

    config_map.add_key_ascii("LoggerGroup", ema.MapEntry.AddEnum, element_list)
    element_list.clear()

    if field_dict_path and enumtype_path:
        dict_elist = (
            ema.ElementList()
            .add_enum("DictionaryType", ema.DictionaryType.File.value)
            .add_ascii("RdmFieldDictionaryFileName", field_dict_path)
            .add_ascii("EnumTypeDefFileName", enumtype_path)
            .complete()
        )
    else:
        if bool(field_dict_path) ^ bool(enumtype_path):
            warnings.warn(
                "Both field_dict_path and enumtype_path need to be defined in the "
                "config file to enable file-based dictionaries."
            )

        dict_elist = (
            ema.ElementList()
            .add_enum(
                "DictionaryType",
                ema.DictionaryType.Channel.value,
            )
            .complete()
        )

    inner_map.add_key_ascii("Dictionary_1", ema.MapEntry.AddEnum, dict_elist).complete()

    element_list.add_map("DictionaryList", inner_map)

    element_list.complete()

    config_map.add_key_ascii("DictionaryGroup", ema.MapEntry.AddEnum, element_list)
    element_list.clear()

    config_map.complete()

    return config_map



# coding: utf8

__all__ = ["BuySell"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class BuySell(Enum):
    """
    The direction of the trade in terms of the deal currency.

    Possible values:
        - Buy
        - Sell
    """

    BUY = "Buy"
    SELL = "Sell"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(BuySell, _BUY_SELL_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_BUY_SELL_VALUES_IN_LOWER_BY_BUY_SELL, some)


_BUY_SELL_VALUES = tuple(t.value for t in BuySell)
_BUY_SELL_VALUES_IN_LOWER_BY_BUY_SELL = {
    name.lower(): item for name, item in BuySell.__members__.items()
}



# coding: utf-8
import os
import re

from setuptools import setup, find_packages

module_file = open("refinitiv/data/__init__.py").read()
metadata = dict(re.findall(r'__([a-z]+)__\s*=\s*"([^"]+)"', module_file))

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="refinitiv-data",
    version=metadata["version"],
    description="Python package for retrieving data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://developers.refinitiv.com/en/api-catalog/refinitiv-data-platform/refinitiv-data-library-for-python",
    author="REFINITIV",
    author_email="",
    license="Apache 2.0",
    data_files=[("", ["LICENSE.md", "CHANGES.txt"])],
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests", "tests_*"]
    ),
    package_data={"": [os.path.join("refinitiv-data.config.json")]},
    zip_safe=False,
    python_requires=">3.6",
    install_requires=[
        "appdirs>=1.4.3",
        "eventemitter>=0.2.0",
        "pyee",
        "httpx>=0.18,<=0.22.0",
        "httpcore<=0.14.5",
        "mysql-connector-python",
        "numpy>=1.11.0",
        "pandas>=1.3.5,<1.4.0",
        "python-configuration>=0.8.2",
        "python-dateutil",
        "requests",
        "scipy",
        "six",
        "tenacity>=8.0,<8.1",
        "watchdog>=0.10.2",
        "websocket-client>=0.58.0,!=1.2.2",
        "pyhumps~=3.0.2",
    ],
)



# coding: utf-8

from . import _json_requests as json_requests
from ._tools import (
    is_string_type,
    check_for_string_or_list_of_strings,
    check_for_string,
    check_for_int,
    get_json_value,
    tz_replacer,
    get_default_session,
)
from .._tools._datetime import to_datetime, get_date_from_today
import pandas as pd
import numpy
from ..errors import RDError

TimeSeries_UDF_endpoint = "TimeSeries"

Calendar_Values = ["native", "tradingdays", "calendardays"]
Corax_Values = ["adjusted", "unadjusted"]


def get_timeseries(
    rics,
    fields=None,
    start_date=None,
    end_date=None,
    interval=None,
    count=None,
    calendar=None,
    corax=None,
    normalize=False,
    raw_output=False,
    debug=False,
):
    """
    Returns historical data on one or several RICs

    Parameters
    ----------
    rics: string or list of strings
        Single RIC or List of RICs to retrieve historical data for

    start_date: string or datetime.datetime or datetime.timedelta
        Starting date and time of the historical range.
        string format is: '%Y-%m-%dT%H:%M:%S'. e.g. '2016-01-20T15:04:05'.
        datetime.timedelta is negative number of day relative to datetime.now().
        Default: datetime.now() + timedelta(-100)
        You can use the helper legacy get_date_from_today, please see the usage in the examples section

    end_date: string or datetime.datetime or datetime.timedelta
        End date and time of the historical range.
        string format could be:
            '%Y-%m-%d' (e.g. '2017-01-20')
            '%Y-%m-%dT%H:%M:%S' (e.g. '2017-01-20T15:04:05')
        datetime.timedelta is negative number of day relative to datetime.now().
        Default: datetime.now()
        You can use the helper legacy get_date_from_today, please see the usage in the examples section

    interval: string
        Data interval.
        Possible values: 'tick', 'minute', 'hour', 'daily', 'weekly', 'monthly', 'quarterly', 'yearly' (Default 'daily')
        Default: 'daily'

    fields: string or list of strings
        Use this parameter to filter the returned fields set.
        Available fields: 'TIMESTAMP', 'VALUE', 'VOLUME', 'HIGH', 'LOW', 'OPEN', 'CLOSE', 'COUNT'
        By default all fields are returned.

    count: int, optional
        Max number of data points retrieved.

    calendar: string, optional
        Possible values: 'native', 'tradingdays', 'calendardays'.

    corax: string, optional
        Possible values: 'adjusted', 'unadjusted'

    normalize: boolean, optional
        if set to True, the legacy will return a normalized data frame with the following columns 'Date','Security','Field'
        If the value of this parameter is False the returned data frame shape will depend on the number of rics and the number of fields in the response
        There are three different shapes
        - One ric and many fields
        - Many rics and one field
        - Many rics and many fields
        Default: False
        Remark: This parameter has a less precedence than the parameter rawOutput i.e. if rawOutput is set to True, the returned data will be the raw data and this parameter will be ignored

    raw_output: boolean, optional
        Set this parameter to True to get the data in json format
        if set to False, the legacy will return a data frame which shape is defined by the parameter normalize
        Default: False

    debug: boolean, optional
        When set to True, the json request and response are printed.
        Default: False

    Raises
    ------
    Exception
        If request fails or if server returns an error
    ValueError
        If a parameter type or value is wrong

    Examples
    --------
    >> import refinitiv.data as ek
    >> ek.set_app_key('set your app key here')
    >> req = ek.get_timeseries(["MSFT.O"], start_date = "2017-02-01T15:04:05",
    >>                          end_date = "2017-02-05T15:04:05", interval="tick")
    >> req = ek.get_timeseries(["MSFT.O"], start_date = "2017-03-01",
    >>                          end_date = "2017-03-10", interval="daily")
    >> req = ek.get_timeseries(["MSFT.O"], start_date = get_date_from_today(150),
    >>                          end_date = get_date_from_today(100), interval="daily")
    """

    logger = get_default_session().logger()

    # set the ric(s) in the payload
    check_for_string_or_list_of_strings(rics, "rics")
    if is_string_type(rics):
        rics = [rics.strip()]

    if type(rics) == list:
        rics = [ric.upper() if ric.islower() else ric for ric in rics]

    # set the field(s) in the payload
    if fields is None or fields == "*":
        fields = ["*"]
    else:
        check_for_string_or_list_of_strings(fields, "fields")
        if is_string_type(fields):
            fields = fields.strip().upper().split()
        else:
            fields = [x.upper() for x in fields]

    if "*" in fields:
        fields = ["*"]
    elif "TIMESTAMP" not in fields:
        fields.append("TIMESTAMP")

    if interval is None:
        interval = "daily"

    # check the interval in the payload
    check_for_string(interval, "interval")

    if start_date is None:
        start_date = get_date_from_today(100)

    if end_date is None:
        end_date = get_date_from_today(0)

    start_date = to_datetime(start_date).isoformat()
    end_date = to_datetime(end_date).isoformat()

    if start_date > end_date:
        error_msg = (
            f"end date ({end_date}) should be after than start date ({start_date})"
        )
        logger.error(error_msg)
        raise ValueError(error_msg)

    payload = {
        "rics": rics,
        "fields": fields,
        "interval": interval,
        "startdate": start_date,
        "enddate": end_date,
    }

    # Add optional parameters

    # set the count in the payload
    if count is not None:
        check_for_int(count, "count")
        payload.update({"count": count})

    # set the calendar in the payload
    if calendar is not None:
        if is_string_type(calendar):
            payload.update({"calendar": calendar})
        else:
            with "calendar must has string type" as error_msg:
                logger.error(error_msg)
                raise ValueError(error_msg)

    # set the corax in the payload
    if corax is not None:
        if is_string_type(corax):
            payload.update({"corax": corax})
        else:
            with "corax must be a string" as error_msg:
                logger.error(error_msg)
                raise ValueError(error_msg)

    response = json_requests.send_json_request(
        TimeSeries_UDF_endpoint, payload, debug=debug
    )
    ts_result = response.json()

    # Catch all errors to raise a warning
    ts_timeserie_data = ts_result["timeseriesData"]
    ts_status_errors = [
        ts_data
        for ts_data in ts_timeserie_data
        if get_json_value(ts_data, "statusCode") == "Error"
    ]

    ts_error_messages = ""
    for ts_status in ts_status_errors:
        ts_error_message = get_json_value(ts_status, "errorMessage")
        ts_error_message = ts_error_message[ts_error_message.find("Description") :]
        ts_instrument = get_json_value(ts_status, "ric")
        ts_error_message = ts_error_message.replace("Description", ts_instrument)
        ts_error_messages += ts_error_message
        ts_error_messages += " | "
        warning_message = "Error with {}".format(ts_error_message)
        logger.warning(warning_message)

    #  if all timeseries are in error, then raise ElektronError with all error messages
    if len(ts_status_errors) == len(ts_timeserie_data):
        logger.error(ts_error_messages)
        raise RDError("Error", message=ts_error_messages)

    if raw_output:
        return ts_result

    data_frame = None
    if normalize:
        data_frame = NormalizedDataFrame_Formatter(ts_result).get_data_frame()
    else:
        data_frame = NiceDataFrame_Formatter(ts_result).get_data_frame()

    if len(data_frame) > 0:
        data_frame = data_frame.fillna(numpy.nan)
    return data_frame


class NormalizedDataFrame_Formatter:
    def __init__(self, json_data):
        self.json_data = json_data

    def get_data_frame(self):
        timeseriesList = self.json_data["timeseriesData"]
        data_frames = []
        for timeseries in timeseriesList:
            ric = timeseries["ric"]
            error_code = timeseries["statusCode"]
            if error_code.lower() == "error":
                continue

            fields = [f["name"] for f in timeseries["fields"]]
            timestamp_index = fields.index("TIMESTAMP")
            fields.pop(
                timestamp_index
            )  # remove timestamp from fields (timestamp is used as index for dataframe)
            datapoints = numpy.array(timeseries["dataPoints"])

            if len(datapoints):
                timestamps = [
                    tz_replacer(value) for value in datapoints[:, timestamp_index]
                ]
                timestamps = numpy.array(
                    timestamps, dtype="datetime64"
                )  # index for dataframe
                # remove timestamp column from numpy array
                datapoints = numpy.delete(datapoints, numpy.s_[timestamp_index], 1)
                fields_count = len(fields)
                column_size = len(datapoints)
                symbol_column = numpy.array([ric] * fields_count * column_size)
                fields_column = numpy.array(fields * column_size)
                values_column = numpy.concatenate(datapoints, axis=0)
                timestamp_column = [
                    [timestamps[i]] * fields_count for i in range(timestamps.size)
                ]
                timestamp_column = numpy.concatenate(timestamp_column, axis=0)
                df = pd.DataFrame(
                    dict(
                        Date=timestamp_column,
                        Security=symbol_column,
                        Field=fields_column,
                        Value=values_column,
                    ),
                    dtype="float",
                )
                if not df.empty:
                    df = df.convert_dtypes()
                data_frames.append(df)
            else:
                data_frames.append(pd.DataFrame([], columns=fields))

        return pd.concat(data_frames)


class NiceDataFrame_Formatter:
    def __init__(self, json_data):
        self.json_data = json_data

    def get_data_frame(self):
        data_frames, rics, fields = self._get_frame_list()
        rics_count = len(rics)
        fields_count = len(fields)
        if rics_count == 0 or fields_count == 0:
            return data_frames
        if rics_count == 1:
            return self._get_frame_1_ric_N_fields(data_frames, rics[0])
        if rics_count > 1 and fields_count == 1:
            return self._get_frame_N_rics_1_field(data_frames, rics, fields[0])
        return self._get_frame_N_rics_N_fields(data_frames, rics, fields)

    def _get_frame_list(self):
        timeseriesList = self.json_data["timeseriesData"]
        data_frames = []
        unique_fields = []
        rics = []
        for timeseries in timeseriesList:
            ric = timeseries["ric"]
            error_code = timeseries["statusCode"]
            if error_code.lower() == "error":
                continue

            rics.append(ric)
            fields = [f["name"] for f in timeseries["fields"]]
            timestamp_index = fields.index("TIMESTAMP")
            fields.pop(
                timestamp_index
            )  # remove timestamp from fields (timestamp is used as index for dataframe)
            unique_fields = fields
            datapoints = numpy.array(timeseries["dataPoints"])
            if len(datapoints):
                timestamps = numpy.array(
                    [tz_replacer(value) for value in datapoints[:, timestamp_index]],
                    dtype="datetime64",
                )  # index for dataframe
                datapoints = numpy.delete(
                    datapoints, numpy.s_[timestamp_index], 1
                )  # remove timestamp column from numpy array
                df = pd.DataFrame(
                    datapoints, columns=fields, index=timestamps, dtype="float"
                )
                if not df.empty:
                    df = df.convert_dtypes()
            else:
                df = pd.DataFrame([], columns=fields)
            if not df.empty:
                df = df.convert_dtypes()
            df.index.name = "Date"
            data_frames.append(df)

        return data_frames, list(rics), list(unique_fields)

    def _get_frame_1_ric_N_fields(self, data_frames, ricName):
        data_frame = pd.concat(data_frames, axis=1)
        if not data_frame.empty:
            data_frame = data_frame.convert_dtypes()
        data_frame.axes[1].name = ricName
        return data_frame

    def _get_frame_N_rics_1_field(self, data_frames, rics, fieldName):
        ric_index = 0
        for df in data_frames:
            ric_name = rics[ric_index]
            df.rename(columns={fieldName: ric_name}, inplace=True)
            ric_index += 1
        data_frame = pd.concat(data_frames, axis=1)
        if not data_frame.empty:
            data_frame = data_frame.convert_dtypes()
        data_frame.axes[1].name = fieldName
        return data_frame

    def _get_frame_N_rics_N_fields(self, data_frames, rics, fields):
        ric_index = 0
        for df in data_frames:
            ric_name = rics[ric_index]
            columns = [(ric_name, f) for f in df.columns]
            df.columns = pd.MultiIndex.from_tuples(columns)
            ric_index += 1
        data_frame = pd.concat(data_frames, axis=1)
        data_frame.axes[1].names = ["Security", "Field"]
        if data_frame.empty:
            return data_frame
        else:
            return data_frame.convert_dtypes()



from typing import Union, Optional
from datetime import timedelta, datetime

from ._tools import _convert_date_time
from ._base_definition import BaseDefinition
from .._data._data_type import DataType
from ..._tools import validate_types


class Definition(BaseDefinition):
    """
    Definition object.

    Parameters
    __________
        fileset_id : str
            The name of the file-set for files to be searched.
            Only exactly matched results are returned.
        file_name : str, optional
            Filter results by file name.
        created_since : str or timedelta or datetime, optional
            Filter results by when the Bucket was created.
        modified_since : str or timedelta or datetime, optional
            Filter results by when the Bucket was last modified.
        skip_token : str, optional
            Used for server-side paging.
        page_size : int, optional
            Used for server-side paging.
            By default 25.

    Methods
    -------
    get_data(session=session)
        Returns a response to the data platform
    get_data_async(session=None)
        Returns a response asynchronously to the data platform

    Examples
    --------
     >>> from refinitiv.data.delivery import cfs
     >>> definition = cfs.files.Definition()
     >>> files = definition.get_data()

     Using get_data_async
     >>> import asyncio
     >>> task = definition.get_data_async()
     >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        fileset_id: str,
        file_name: Optional[str] = None,
        created_since: Union[str, datetime, timedelta] = None,
        modified_since: Union[str, datetime, timedelta] = None,
        skip_token: Optional[str] = None,
        page_size: Optional[int] = 25,
    ):
        validate_types(page_size, [int, type(None)], "page_size")
        created_since = _convert_date_time(created_since)
        modified_since = _convert_date_time(modified_since)
        super().__init__(
            data_type=DataType.CFS_FILES,
            fileset_id=fileset_id,
            file_name=file_name,
            created_since=created_since,
            modified_since=modified_since,
            skip_token=skip_token,
            page_size=page_size,
        )



# coding: utf8

from typing import Optional

from .._instrument_definition import InstrumentDefinition
from ._enums import (
    AdjustInterestToPaymentDate,
    IndexResetType,
    Frequency,
    DateRollingConvention,
    DayCountBasis,
    StubRule,
    BuySell,
    BusinessDayConvention,
)
from ._models import (
    AmortizationItem,
    BarrierDefinitionElement,
)


class CapFloorInstrumentDefinition(InstrumentDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    instrument_tag : str, optional
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
    start_date : str, optional
        The option start date
    end_date : str, optional
        The maturity date of the CapFloor
    tenor : str, optional
        Tenor of the option
    notional_ccy : str, optional
        The ISO code of the notional currency. Mandatory if instrument code or
        instrument style has not been defined. In case an instrument code/style has been
        defined, value may comes from the reference data.
    notional_amount : float, optional
        The notional amount of the leg at the period start date. Optional. By default
        1,000,000 is used.
    index_name : str, optional
        The name of the floating rate index.
    index_tenor : str, optional
        The period code that represents the maturity of the floating rate index.
        Mandatory when the leg is float.
    interest_payment_frequency : Frequency, optional
        The frequency of the interest payments. Optional if an instrument code/style
        have been defined : in that case, value comes from reference data. Otherwise, it
        is mandatory.
    interest_calculation_method : DayCountBasis, optional
        The Day Count Basis method used to calculate the coupon interest payments.
        Mandatory.
    payment_business_day_convention : BusinessDayConvention, optional
        The method to adjust dates to a working day. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. In case an instrument code/style has been defined,
          value comes from bond reference data. Otherwise 'ModifiedFollowing' is used.
    payment_roll_convention : DateRollingConvention, optional
        The method to adjust payment dates whn they fall at the end of the month (28th
        of February, 30th, 31st). The possible values are:
        - Last (For setting the calculated date to the last working day),
        - Same (For setting the calculated date to the same day . In this latter case,
          the date may be moved according to the date moving convention if it is a
          non-working day),
        - Last28 (For setting the calculated date to the last working day. 28FEB being
          always considered as the last working day),
        - Same28 (For setting the calculated date to the same day .28FEB being always
          considered as the last working day). Optional. By default 'SameDay' is used.
    index_reset_frequency : Frequency, optional
        The reset frequency in case the leg Type is Float. Optional. By default the
        IndexTenor is used.
    index_reset_type : IndexResetType, optional
        A flag that indicates if the floating rate index is reset before the coupon
        period starts or at the end of the coupon period. The possible values are:
        - InAdvance (resets the index before the start of the interest period),
        - InArrears (resets the index at the end of the interest period). Optional. By
          default 'InAdvance' is used.
    index_fixing_lag : int, optional
        Defines the positive number of working days between the fixing date and the
        start of the coupon period ('InAdvance') or the end of the coupon period
        ('InArrears'). Optional. By default 0 is used.
    amortization_schedule : AmortizationItem, optional
        Definition of amortizations
    payment_business_days : str, optional
        A list of coma-separated calendar codes to adjust dates (e.g. 'EMU' or 'USA').
        Optional. By default the calendar associated to NotionalCcy is used.
    adjust_interest_to_payment_date : AdjustInterestToPaymentDate, optional
        A flag that indicates if the coupon dates are adjusted to the payment dates.
        Optional. By default 'false' is used.
    stub_rule : StubRule, optional
        The rule that defines whether coupon roll dates are aligned on the  maturity or
        the issue date. The possible values are:
        - ShortFirstProRata (to create a short period between the start date and the
          first coupon date, and pay a smaller amount of interest for the short
          period.All coupon dates are calculated backward from the maturity date),
        - ShortFirstFull (to create a short period between the start date and the first
          coupon date, and pay a regular coupon on the first coupon date. All coupon
          dates are calculated backward from the maturity date),
        - LongFirstFull (to create a long period between the start date and the second
          coupon date, and pay a regular coupon on the second coupon date. All coupon
          dates are calculated backward from the maturity date),
        - ShortLastProRata (to create a short period between the last payment date and
          maturity, and pay a smaller amount of interest for the short period. All
          coupon dates are calculated forward from the start date). This property may
          also be used in conjunction with firstRegularPaymentDate and
          lastRegularPaymentDate; in that case the following values can be defined:
        - Issue (all dates are aligned on the issue date),
        - Maturity (all dates are aligned on the maturity date). Optional. By default
          'Maturity' is used.
    barrier_definition : BarrierDefinitionElement, optional

    buy_sell : BuySell, optional
        The side of the deal. Possible values:
        - Buy
        - Sell
    annualized_rebate : bool, optional

    cap_digital_payout_percent : float, optional

    cap_strike_percent : float, optional
        Cap leg strike expressed in %
    cms_template : str, optional
        A reference to a common swap contract that represents the underlying swap in
        case of a Constant Maturity Swap contract (CMS). Mandatory for CMS contract.
    floor_digital_payout_percent : float, optional

    floor_strike_percent : float, optional
        Floor leg strike expressed in %
    index_fixing_ric : str, optional
        The RIC that carries the fixing value. This value overrides the RIC associated
        by default with the IndexName and IndexTenor. Optional.
    """

    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        notional_ccy: Optional[str] = None,
        notional_amount: Optional[float] = None,
        index_name: Optional[str] = None,
        index_tenor: Optional[str] = None,
        interest_payment_frequency: Optional[Frequency] = None,
        interest_calculation_method: Optional[DayCountBasis] = None,
        payment_business_day_convention: Optional[BusinessDayConvention] = None,
        payment_roll_convention: Optional[DateRollingConvention] = None,
        index_reset_frequency: Optional[Frequency] = None,
        index_reset_type: Optional[IndexResetType] = None,
        index_fixing_lag: Optional[int] = None,
        amortization_schedule: Optional[AmortizationItem] = None,
        payment_business_days: Optional[str] = None,
        adjust_interest_to_payment_date: Optional[AdjustInterestToPaymentDate] = None,
        stub_rule: Optional[StubRule] = None,
        barrier_definition: Optional[BarrierDefinitionElement] = None,
        buy_sell: Optional[BuySell] = None,
        annualized_rebate: Optional[bool] = None,
        cap_digital_payout_percent: Optional[float] = None,
        cap_strike_percent: Optional[float] = None,
        cms_template: Optional[str] = None,
        floor_digital_payout_percent: Optional[float] = None,
        floor_strike_percent: Optional[float] = None,
        index_fixing_ric: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.instrument_tag = instrument_tag
        self.start_date = start_date
        self.end_date = end_date
        self.tenor = tenor
        self.notional_ccy = notional_ccy
        self.notional_amount = notional_amount
        self.index_name = index_name
        self.index_tenor = index_tenor
        self.interest_payment_frequency = interest_payment_frequency
        self.interest_calculation_method = interest_calculation_method
        self.payment_business_day_convention = payment_business_day_convention
        self.payment_roll_convention = payment_roll_convention
        self.index_reset_frequency = index_reset_frequency
        self.index_reset_type = index_reset_type
        self.index_fixing_lag = index_fixing_lag
        self.amortization_schedule = amortization_schedule
        self.payment_business_days = payment_business_days
        self.adjust_interest_to_payment_date = adjust_interest_to_payment_date
        self.stub_rule = stub_rule
        self.barrier_definition = barrier_definition
        self.buy_sell = buy_sell
        self.annualized_rebate = annualized_rebate
        self.cap_digital_payout_percent = cap_digital_payout_percent
        self.cap_strike_percent = cap_strike_percent
        self.cms_template = cms_template
        self.floor_digital_payout_percent = floor_digital_payout_percent
        self.floor_strike_percent = floor_strike_percent
        self.index_fixing_ric = index_fixing_ric

    @classmethod
    def get_instrument_type(cls):
        return "CapFloor"

    @property
    def adjust_interest_to_payment_date(self):
        """
        A flag that indicates if the coupon dates are adjusted to the payment dates.
        Optional. By default 'false' is used.
        :return: enum AdjustInterestToPaymentDate
        """
        return self._get_enum_parameter(
            AdjustInterestToPaymentDate, "adjustInterestToPaymentDate"
        )

    @adjust_interest_to_payment_date.setter
    def adjust_interest_to_payment_date(self, value):
        self._set_enum_parameter(
            AdjustInterestToPaymentDate, "adjustInterestToPaymentDate", value
        )

    @property
    def amortization_schedule(self):
        """
        Definition of amortizations
        :return: list AmortizationItem
        """
        return self._get_list_parameter(AmortizationItem, "amortizationSchedule")

    @amortization_schedule.setter
    def amortization_schedule(self, value):
        self._set_list_parameter(AmortizationItem, "amortizationSchedule", value)

    @property
    def barrier_definition(self):
        """
        :return: object BarrierDefinitionElement
        """
        return self._get_object_parameter(BarrierDefinitionElement, "barrierDefinition")

    @barrier_definition.setter
    def barrier_definition(self, value):
        self._set_object_parameter(BarrierDefinitionElement, "barrierDefinition", value)

    @property
    def buy_sell(self):
        """
        The side of the deal. Possible values:
        - Buy
        - Sell
        :return: enum BuySell
        """
        return self._get_enum_parameter(BuySell, "buySell")

    @buy_sell.setter
    def buy_sell(self, value):
        self._set_enum_parameter(BuySell, "buySell", value)

    @property
    def index_reset_frequency(self):
        """
        The reset frequency in case the leg Type is Float. Optional. By default the
        IndexTenor is used.
        :return: enum Frequency
        """
        return self._get_enum_parameter(Frequency, "indexResetFrequency")

    @index_reset_frequency.setter
    def index_reset_frequency(self, value):
        self._set_enum_parameter(Frequency, "indexResetFrequency", value)

    @property
    def index_reset_type(self):
        """
        A flag that indicates if the floating rate index is reset before the coupon
        period starts or at the end of the coupon period. The possible values are:
        - InAdvance (resets the index before the start of the interest period),
        - InArrears (resets the index at the end of the interest period). Optional. By
          default 'InAdvance' is used.
        :return: enum IndexResetType
        """
        return self._get_enum_parameter(IndexResetType, "indexResetType")

    @index_reset_type.setter
    def index_reset_type(self, value):
        self._set_enum_parameter(IndexResetType, "indexResetType", value)

    @property
    def interest_calculation_method(self):
        """
        The Day Count Basis method used to calculate the coupon interest payments.
        Mandatory.
        :return: enum DayCountBasis
        """
        return self._get_enum_parameter(DayCountBasis, "interestCalculationMethod")

    @interest_calculation_method.setter
    def interest_calculation_method(self, value):
        self._set_enum_parameter(DayCountBasis, "interestCalculationMethod", value)

    @property
    def interest_payment_frequency(self):
        """
        The frequency of the interest payments. Optional if an instrument code/style
        have been defined : in that case, value comes from reference data. Otherwise, it
        is mandatory.
        :return: enum Frequency
        """
        return self._get_enum_parameter(Frequency, "interestPaymentFrequency")

    @interest_payment_frequency.setter
    def interest_payment_frequency(self, value):
        self._set_enum_parameter(Frequency, "interestPaymentFrequency", value)

    @property
    def payment_business_day_convention(self):
        """
        The method to adjust dates to a working day. The possible values are:
        - ModifiedFollowing (adjusts dates according to the Modified Following
          convention - next business day unless is it goes into the next month,
          preceeding is used in that  case),
        - NextBusinessDay (adjusts dates according to the Following convention - Next
          Business Day),
        - PreviousBusinessDay (adjusts dates  according to the Preceeding convention -
          Previous Business Day),
        - NoMoving (does not adjust dates),
        - BbswModifiedFollowing (adjusts dates  according to the BBSW Modified Following
          convention). Optional. In case an instrument code/style has been defined,
          value comes from bond reference data. Otherwise 'ModifiedFollowing' is used.
        :return: enum BusinessDayConvention
        """
        return self._get_enum_parameter(
            BusinessDayConvention, "paymentBusinessDayConvention"
        )

    @payment_business_day_convention.setter
    def payment_business_day_convention(self, value):
        self._set_enum_parameter(
            BusinessDayConvention, "paymentBusinessDayConvention", value
        )

    @property
    def payment_roll_convention(self):
        """
        The method to adjust payment dates whn they fall at the end of the month (28th
        of February, 30th, 31st). The possible values are:
        - Last (For setting the calculated date to the last working day),
        - Same (For setting the calculated date to the same day . In this latter case,
          the date may be moved according to the date moving convention if it is a
          non-working day),
        - Last28 (For setting the calculated date to the last working day. 28FEB being
          always considered as the last working day),
        - Same28 (For setting the calculated date to the same day .28FEB being always
          considered as the last working day). Optional. By default 'SameDay' is used.
        :return: enum DateRollingConvention
        """
        return self._get_enum_parameter(DateRollingConvention, "paymentRollConvention")

    @payment_roll_convention.setter
    def payment_roll_convention(self, value):
        self._set_enum_parameter(DateRollingConvention, "paymentRollConvention", value)

    @property
    def stub_rule(self):
        """
        The rule that defines whether coupon roll dates are aligned on the  maturity or
        the issue date. The possible values are:
        - ShortFirstProRata (to create a short period between the start date and the
          first coupon date, and pay a smaller amount of interest for the short
          period.All coupon dates are calculated backward from the maturity date),
        - ShortFirstFull (to create a short period between the start date and the first
          coupon date, and pay a regular coupon on the first coupon date. All coupon
          dates are calculated backward from the maturity date),
        - LongFirstFull (to create a long period between the start date and the second
          coupon date, and pay a regular coupon on the second coupon date. All coupon
          dates are calculated backward from the maturity date),
        - ShortLastProRata (to create a short period between the last payment date and
          maturity, and pay a smaller amount of interest for the short period. All
          coupon dates are calculated forward from the start date). This property may
          also be used in conjunction with firstRegularPaymentDate and
          lastRegularPaymentDate; in that case the following values can be defined:
        - Issue (all dates are aligned on the issue date),
        - Maturity (all dates are aligned on the maturity date). Optional. By default
          'Maturity' is used.
        :return: enum StubRule
        """
        return self._get_enum_parameter(StubRule, "stubRule")

    @stub_rule.setter
    def stub_rule(self, value):
        self._set_enum_parameter(StubRule, "stubRule", value)

    @property
    def annualized_rebate(self):
        """
        :return: bool
        """
        return self._get_parameter("annualizedRebate")

    @annualized_rebate.setter
    def annualized_rebate(self, value):
        self._set_parameter("annualizedRebate", value)

    @property
    def cap_digital_payout_percent(self):
        """
        :return: float
        """
        return self._get_parameter("capDigitalPayoutPercent")

    @cap_digital_payout_percent.setter
    def cap_digital_payout_percent(self, value):
        self._set_parameter("capDigitalPayoutPercent", value)

    @property
    def cap_strike_percent(self):
        """
        Cap leg strike expressed in %
        :return: float
        """
        return self._get_parameter("capStrikePercent")

    @cap_strike_percent.setter
    def cap_strike_percent(self, value):
        self._set_parameter("capStrikePercent", value)

    @property
    def cms_template(self):
        """
        A reference to a common swap contract that represents the underlying swap in
        case of a Constant Maturity Swap contract (CMS). Mandatory for CMS contract.
        :return: str
        """
        return self._get_parameter("cmsTemplate")

    @cms_template.setter
    def cms_template(self, value):
        self._set_parameter("cmsTemplate", value)

    @property
    def end_date(self):
        """
        The maturity date of the CapFloor
        :return: str
        """
        return self._get_parameter("endDate")

    @end_date.setter
    def end_date(self, value):
        self._set_parameter("endDate", value)

    @property
    def floor_digital_payout_percent(self):
        """
        :return: float
        """
        return self._get_parameter("floorDigitalPayoutPercent")

    @floor_digital_payout_percent.setter
    def floor_digital_payout_percent(self, value):
        self._set_parameter("floorDigitalPayoutPercent", value)

    @property
    def floor_strike_percent(self):
        """
        Floor leg strike expressed in %
        :return: float
        """
        return self._get_parameter("floorStrikePercent")

    @floor_strike_percent.setter
    def floor_strike_percent(self, value):
        self._set_parameter("floorStrikePercent", value)

    @property
    def index_fixing_lag(self):
        """
        Defines the positive number of working days between the fixing date and the
        start of the coupon period ('InAdvance') or the end of the coupon period
        ('InArrears'). Optional. By default 0 is used.
        :return: int
        """
        return self._get_parameter("indexFixingLag")

    @index_fixing_lag.setter
    def index_fixing_lag(self, value):
        self._set_parameter("indexFixingLag", value)

    @property
    def index_fixing_ric(self):
        """
        The RIC that carries the fixing value. This value overrides the RIC associated
        by default with the IndexName and IndexTenor. Optional.
        :return: str
        """
        return self._get_parameter("indexFixingRic")

    @index_fixing_ric.setter
    def index_fixing_ric(self, value):
        self._set_parameter("indexFixingRic", value)

    @property
    def index_name(self):
        """
        The name of the floating rate index.
        :return: str
        """
        return self._get_parameter("indexName")

    @index_name.setter
    def index_name(self, value):
        self._set_parameter("indexName", value)

    @property
    def index_tenor(self):
        """
        The period code that represents the maturity of the floating rate index.
        Mandatory when the leg is float.
        :return: str
        """
        return self._get_parameter("indexTenor")

    @index_tenor.setter
    def index_tenor(self, value):
        self._set_parameter("indexTenor", value)

    @property
    def instrument_tag(self):
        """
        User defined string to identify the instrument.It can be used to link output
        results to the instrument definition. Only alphabetic, numeric and '- _.#=@'
        characters are supported. Optional.
        :return: str
        """
        return self._get_parameter("instrumentTag")

    @instrument_tag.setter
    def instrument_tag(self, value):
        self._set_parameter("instrumentTag", value)

    @property
    def notional_amount(self):
        """
        The notional amount of the leg at the period start date. Optional. By default
        1,000,000 is used.
        :return: float
        """
        return self._get_parameter("notionalAmount")

    @notional_amount.setter
    def notional_amount(self, value):
        self._set_parameter("notionalAmount", value)

    @property
    def notional_ccy(self):
        """
        The ISO code of the notional currency. Mandatory if instrument code or
        instrument style has not been defined. In case an instrument code/style has been
        defined, value may comes from the reference data.
        :return: str
        """
        return self._get_parameter("notionalCcy")

    @notional_ccy.setter
    def notional_ccy(self, value):
        self._set_parameter("notionalCcy", value)

    @property
    def payment_business_days(self):
        """
        A list of coma-separated calendar codes to adjust dates (e.g. 'EMU' or 'USA').
        Optional. By default the calendar associated to NotionalCcy is used.
        :return: str
        """
        return self._get_parameter("paymentBusinessDays")

    @payment_business_days.setter
    def payment_business_days(self, value):
        self._set_parameter("paymentBusinessDays", value)

    @property
    def start_date(self):
        """
        The option start date
        :return: str
        """
        return self._get_parameter("startDate")

    @start_date.setter
    def start_date(self, value):
        self._set_parameter("startDate", value)

    @property
    def tenor(self):
        """
        Tenor of the option
        :return: str
        """
        return self._get_parameter("tenor")

    @tenor.setter
    def tenor(self, value):
        self._set_parameter("tenor", value)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.246815

__all__ = ["NotionalExchange"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class NotionalExchange(Enum):
    BOTH = "Both"
    END = "End"
    END_ADJUSTMENT = "EndAdjustment"
    NONE = "None"
    START = "Start"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(NotionalExchange, _NOTIONAL_EXCHANGE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_NOTIONAL_EXCHANGE_VALUES_IN_LOWER_BY_NOTIONAL_EXCHANGE, some)


_NOTIONAL_EXCHANGE_VALUES = tuple(t.value for t in NotionalExchange)
_NOTIONAL_EXCHANGE_VALUES_IN_LOWER_BY_NOTIONAL_EXCHANGE = {
    name.lower(): item for name, item in NotionalExchange.__members__.items()
}



import re
from typing import Iterable, Union, List

import pandas as pd
from pandas import DataFrame, MultiIndex
from pandas.core.tools.datetimes import DatetimeScalarOrArrayConvertible


def convert_df_columns_to_datetime(
    df: pd.DataFrame, pattern: str, utc: bool = None, delete_tz: bool = False
) -> pd.DataFrame:
    for i in df.columns:
        if pattern in i:
            df[i] = pd.to_datetime(df[i], utc=utc)

            if delete_tz:
                df[i] = df[i].dt.tz_localize(None)

    return df


def convert_df_columns_to_datetime_use_re_compile(
    df: pd.DataFrame, re_compile: re.compile
) -> pd.DataFrame:
    for column_name in df.columns:
        if re_compile.search(column_name):
            convert_column_df_to_datetime(df, column_name)

    return df


def convert_str_to_datetime(s: str) -> DatetimeScalarOrArrayConvertible:
    date = pd.to_datetime(s, utc=True)
    date = date.tz_localize(None)
    return date


def convert_column_df_to_datetime(
    df: pd.DataFrame, column_name: Union[int, str]
) -> pd.DataFrame:
    df[column_name] = pd.to_datetime(df[column_name], utc=True)
    df[column_name] = df[column_name].dt.tz_localize(None)

    return df


def convert_items_df_to_datetime(
    items: DatetimeScalarOrArrayConvertible,
) -> DatetimeScalarOrArrayConvertible:
    date_time_items = pd.to_datetime(items, utc=True)
    date_time_items = date_time_items.tz_localize(None)

    return date_time_items


def sort_df_by_universe(df: DataFrame, universe: List[str]) -> DataFrame:
    length = len(universe)

    if length == 1:
        return df

    columns = df.columns

    def make_getidx():
        get_index = universe.index
        if isinstance(columns, MultiIndex):

            def geti(i):
                return i[0]

        else:

            def geti(i):
                return i

        def inner(i):
            try:
                index = get_index(geti(i))
            except ValueError:
                index = length
            return index

        return inner

    getidx = make_getidx()
    # [3, 0, 2, 1]
    curr_order = [getidx(col) for col in columns]
    # [0, 1, 2, 3]
    expected_order = list(range(length))
    if curr_order != expected_order:
        sorted_columns = (col for _, col in sorted(zip(curr_order, columns)))
        df = df.reindex(columns=sorted_columns)
    return df



# coding: utf8

__all__ = ["EtiInputVolatilityType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class EtiInputVolatilityType(Enum):
    IMPLIED = "Implied"
    QUOTED = "Quoted"
    SETTLE = "Settle"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            EtiInputVolatilityType, _ETI_INPUT_VOLATILITY_TYPE_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _ETI_INPUT_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_ETI_INPUT_VOLATILITY_TYPE,
            some,
        )


_ETI_INPUT_VOLATILITY_TYPE_VALUES = tuple(t.value for t in EtiInputVolatilityType)
_ETI_INPUT_VOLATILITY_TYPE_VALUES_IN_LOWER_BY_ETI_INPUT_VOLATILITY_TYPE = {
    name.lower(): item for name, item in EtiInputVolatilityType.__members__.items()
}



# coding: utf8

__all__ = ["SurfaceOutput"]

from ..._object_definition import ObjectDefinition
from ._volatility_surface_point import VolatilitySurfacePoint
from .._enums import Format
from .._surface_types import OptFormat, OptVolatilitySurfacePoints
from ...._types import OptStrings, OptInt
from ....._tools import create_repr


class SurfaceOutput(ObjectDefinition):
    """

    Parameters
    ---------
    data_points : list of VolatilitySurfacePoint, optional
        Specifies the list of specific data points to be returned
    format : Format, option
        Specifies whether the calculated volatilities are returned as a list or a matrix
    x_values : list of str, optional
        Specifies a list of discrete values for the x-axis
    y_values : list of str, optional
        Specifies a list of discrete values for the y-axis
    z_values : list of str, optional
        Specifies a list of discrete values for the z-axis
    x_point_count : int, optional
        Specifies the number of values that will be generated along the x-axis.
        These values will distributed depending on the available input data and the type
        of volatility
    y_point_count : int, optional
        Specifies the number of values that will be generated along the y-axis.
        These values will distributed depending on the available input data and the type
        of volatility
    z_point_count : int, optional
        Specifies the number of values that will be generated along the z-axis.
        These values will distributed depending on the available input data and the type
        of volatility

    Examples
    -------
    >>> from refinitiv.data.content.ipa.surfaces.surface_fx import SurfaceLayout
    >>> from refinitiv.data.content.ipa.surfaces import surface_fx
    >>> SurfaceLayout(format=surface_fx.Format.MATRIX)
    """

    def __init__(
        self,
        data_points: OptVolatilitySurfacePoints = None,
        format: OptFormat = None,
        x_values: OptStrings = None,
        y_values: OptStrings = None,
        z_values: OptStrings = None,
        x_point_count: OptInt = None,
        y_point_count: OptInt = None,
        z_point_count: OptInt = None,
    ):
        super().__init__()
        self.data_points = data_points
        self.format = format
        self.x_values = x_values
        self.y_values = y_values
        self.z_values = z_values
        self.x_point_count = x_point_count
        self.y_point_count = y_point_count
        self.z_point_count = z_point_count

    def __repr__(self):
        return create_repr(
            self,
            middle_path="surfaces.surface_fx",
            class_name="SurfaceLayout",
        )

    @property
    def data_points(self):
        """
        Specifies the list of specific data points to be returned.
        :return: list VolatilitySurfacePoint
        """
        return self._get_list_parameter(VolatilitySurfacePoint, "dataPoints")

    @data_points.setter
    def data_points(self, value):
        self._set_list_parameter(VolatilitySurfacePoint, "dataPoints", value)

    @property
    def format(self):
        """
        Specifies whether the calculated volatilities are returned as a list or a matrix.
        :return: enum Format
        """
        return self._get_enum_parameter(Format, "format")

    @format.setter
    def format(self, value):
        self._set_enum_parameter(Format, "format", value)

    @property
    def x_values(self):
        """
        Specifies a list of discrete values for the x-axis.
        :return: list string
        """
        return self._get_list_parameter(str, "xValues")

    @x_values.setter
    def x_values(self, value):
        self._set_list_parameter(str, "xValues", value)

    @property
    def y_values(self):
        """
        Specifies a list of discrete values for the y-axis.
        :return: list string
        """
        return self._get_list_parameter(str, "yValues")

    @y_values.setter
    def y_values(self, value):
        self._set_list_parameter(str, "yValues", value)

    @property
    def z_values(self):
        """
        Specifies a list of discrete values for the z-axis.
        :return: list string
        """
        return self._get_list_parameter(str, "zValues")

    @z_values.setter
    def z_values(self, value):
        self._set_list_parameter(str, "zValues", value)

    @property
    def x_point_count(self):
        """
        Specifies the number of values that will be generated along the x-axis.
        These values will distributed depending on the available input data and the type of volatility.
        :return: int
        """
        return self._get_parameter("xPointCount")

    @x_point_count.setter
    def x_point_count(self, value):
        self._set_parameter("xPointCount", value)

    @property
    def y_point_count(self):
        """
        Specifies the number of values that will be generated along the y-axis.
        These values will distributed depending on the available input data and the type of volatility.
        :return: int
        """
        return self._get_parameter("yPointCount")

    @y_point_count.setter
    def y_point_count(self, value):
        self._set_parameter("yPointCount", value)

    @property
    def z_point_count(self):
        """
        Specifies the number of values that will be generated along the z-axis.
        These values will distributed depending on the available input data and the type of volatility.
        :return: int
        """
        return self._get_parameter("zPointCount")

    @z_point_count.setter
    def z_point_count(self, value):
        self._set_parameter("zPointCount", value)



# coding: utf8
from typing import List, Callable

import numpy as np
import pandas as pd
from numpy import iterable

from ._models._curve import Curve, ForwardCurve, ZcCurve
from .._content_provider import CurvesAndSurfacesRequestFactory, get_type_by_axis
from ..._content_type import ContentType
from ...._tools._ipa_validate import do_all_elements_have_error
from ...._tools._dataframe import convert_df_columns_to_datetime
from ....delivery._data._data_provider import ContentValidator, ValidatorContainer
from ....delivery._data._data_provider import DataProvider, ResponseFactory, Data


# ---------------------------------------------------------------------------
#   ContentValidator
# ---------------------------------------------------------------------------


class CurvesContentValidator(ContentValidator):
    def validate(self, data, *args, **kwargs):
        is_valid = True
        content_data = data.get("content_data")
        raw_curves = content_data.get("data")

        if not raw_curves:
            is_valid = False
            data["error_code"] = 1
            data["error_message"] = "Content data is None"

        elif isinstance(raw_curves, list):
            is_valid = not do_all_elements_have_error(data, raw_curves)

        return is_valid


# ---------------------------------------------------------------------------
#   Content data
# ---------------------------------------------------------------------------


def zc_curves_build_df(raw, **kwargs):
    """

    Parameters
    ----------
    raw : dict

    Returns
    -------
    DataFrame
    >>> raw
    ... {
    ...     "data": [
    ...         {
    ...             "curveTag": "TAG",
    ...             "error": {
    ...                 "id": "9fef13f4-6d11-4d71-a388-824ddcc8a95a/9fef13f4-6d11-4d71-a388-824ddcc8a95a",
    ...                 "code": "QPS-Curves.7",
    ...                 "message": "The service failed to find the curve definition",
    ...             },
    ...         },
    ...         {
    ...             "curveParameters": {
    ...                 "extrapolationMode": "None",
    ...                 "interpolationMode": "CubicDiscount",
    ...                 "interestCalculationMethod": "Dcb_Actual_Actual",
    ...                 "priceSide": "Mid",
    ...                 "calendarAdjustment": "Calendar",
    ...                 "calendars": ["EMU_FI"],
    ...                 "compoundingType": "Compounded",
    ...                 "useMultiDimensionalSolver": True,
    ...                 "useConvexityAdjustment": True,
    ...                 "useSteps": False,
    ...                 "convexityAdjustment": {
    ...                     "meanReversionPercent": 3.9012,
    ...                     "volatilityPercent": 0.863,
    ...                 },
    ...                 "valuationDate": "2022-02-09",
    ...             },
    ...             "curveDefinition": {
    ...                 "availableTenors": ["OIS", "1M", "3M", "6M", "1Y"],
    ...                 "availableDiscountingTenors": ["OIS", "1M", "3M", "6M", "1Y"],
    ...                 "currency": "EUR",
    ...                 "mainConstituentAssetClass": "Swap",
    ...                 "riskType": "InterestRate",
    ...                 "indexName": "EURIBOR",
    ...                 "source": "Refinitiv",
    ...                 "name": "EUR EURIBOR Swap ZC Curve",
    ...                 "id": "9d619112-9ab3-45c9-b83c-eb04cbec382e",
    ...                 "discountingTenor": "OIS",
    ...                 "ignoreExistingDefinition": False,
    ...                 "owner": "Refinitiv",
    ...                 "indexTenors": ["OIS", "1M", "3M", "6M", "1Y"],
    ...             },
    ...             "curves": {
    ...                 "OIS": {
    ...                     "curvePoints": [
    ...                         {
    ...                             "endDate": "2022-02-09",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0,
    ...                             "ratePercent": -0.49456799906775206,
    ...                             "tenor": "0D",
    ...                         },
    ...                         {
    ...                             "endDate": "2022-02-10",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0000135835178428,
    ...                             "ratePercent": -0.49456799906775206,
    ...                             "tenor": "ON",
    ...                             "instruments": [{"instrumentCode": "EUROSTR="}],
    ...                         },
    ...                     ],
    ...                     "isDiscountCurve": True,
    ...                 },
    ...                 "1M": {
    ...                     "curvePoints": [
    ...                         {
    ...                             "endDate": "2022-02-09",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0,
    ...                             "ratePercent": -0.5560912053716005,
    ...                             "tenor": "0D",
    ...                         },
    ...                         {
    ...                             "endDate": "2022-02-10",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0000152780111917,
    ...                             "ratePercent": -0.5560912053716005,
    ...                             "tenor": "ON",
    ...                             "instruments": [{"instrumentCode": "EUROND="}],
    ...                         },
    ...                     ],
    ...                     "isDiscountCurve": False,
    ...                 },
    ...                 "3M": {
    ...                     "curvePoints": [
    ...                         {
    ...                             "endDate": "2022-02-09",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0,
    ...                             "ratePercent": -0.5560912053716005,
    ...                             "tenor": "0D",
    ...                         },
    ...                         {
    ...                             "endDate": "2022-02-10",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0000152780111917,
    ...                             "ratePercent": -0.5560912053716005,
    ...                             "tenor": "ON",
    ...                             "instruments": [{"instrumentCode": "EUROND="}],
    ...                         },
    ...                     ],
    ...                     "isDiscountCurve": False,
    ...                 },
    ...                 "6M": {
    ...                     "curvePoints": [
    ...                         {
    ...                             "endDate": "2022-02-09",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0,
    ...                             "ratePercent": -0.5560912053716005,
    ...                             "tenor": "0D",
    ...                         },
    ...                         {
    ...                             "endDate": "2022-02-10",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0000152780111917,
    ...                             "ratePercent": -0.5560912053716005,
    ...                             "tenor": "ON",
    ...                             "instruments": [{"instrumentCode": "EUROND="}],
    ...                         },
    ...                     ],
    ...                     "isDiscountCurve": False,
    ...                 },
    ...                 "1Y": {
    ...                     "curvePoints": [
    ...                         {
    ...                             "endDate": "2022-02-09",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0,
    ...                             "ratePercent": -0.5560912053716005,
    ...                             "tenor": "0D",
    ...                         },
    ...                         {
    ...                             "endDate": "2022-02-10",
    ...                             "startDate": "2022-02-09",
    ...                             "discountFactor": 1.0000152780111917,
    ...                             "ratePercent": -0.5560912053716005,
    ...                             "tenor": "ON",
    ...                             "instruments": [{"instrumentCode": "EUROND="}],
    ...                         },
    ...                     ],
    ...                     "isDiscountCurve": False,
    ...                 },
    ...             },
    ...         },
    ...     ]
    ... }
    """
    datas = raw.get("data", [])
    datas = datas or []
    dfs = []

    for data in datas:

        error = data.get("error")
        if error:
            continue

        curves = data.get("curves")
        for value in curves.values():
            curve_points = value.get("curvePoints")

            d = {}
            for curve_point in curve_points:
                for key, value in curve_point.items():
                    values = d.setdefault(key, [])
                    values.append(value)

            d.pop("instruments", None)

            df = pd.DataFrame(d)
            df = df.convert_dtypes()
            dfs.append(df)

    df = pd.concat(dfs, ignore_index=True)
    df = convert_df_columns_to_datetime(df, "Date", utc=True, delete_tz=True)
    return df


def forward_curve_build_df(raw, **kwargs):
    """

    Parameters
    ----------
    raw : dict

    Returns
    -------
    DataFrame

    Examples
    -------
    >>> raw
    ... {
    ...     "data": [
    ...         {
    ...             "error": {
    ...                 "id": "b6f9797d-72c8-4baa-84eb-6a079fc40ec5/b6f9797d-72c8-4baa-84eb-6a079fc40ec5",
    ...                 "code": "QPS-Curves.6",
    ...                 "message": "Invalid input: curveDefinition is missing",
    ...             }
    ...         },
    ...         {
    ...             "curveTag": "test_curve",
    ...             "curveParameters": {
    ...                 "interestCalculationMethod": "Dcb_Actual_Actual",
    ...                 "priceSide": "Mid",
    ...                 "calendarAdjustment": "Calendar",
    ...                 "calendars": ["EMU_FI"],
    ...                 "compoundingType": "Compounded",
    ...                 "useConvexityAdjustment": True,
    ...                 "useSteps": False,
    ...                 "valuationDate": "2022-02-09",
    ...             },
    ...             "curveDefinition": {
    ...                 "availableTenors": ["OIS", "1M", "3M", "6M", "1Y"],
    ...                 "availableDiscountingTenors": ["OIS", "1M", "3M", "6M", "1Y"],
    ...                 "currency": "EUR",
    ...                 "mainConstituentAssetClass": "Swap",
    ...                 "riskType": "InterestRate",
    ...                 "indexName": "EURIBOR",
    ...                 "source": "Refinitiv",
    ...                 "name": "EUR EURIBOR Swap ZC Curve",
    ...                 "id": "9d619112-9ab3-45c9-b83c-eb04cbec382e",
    ...                 "discountingTenor": "OIS",
    ...                 "ignoreExistingDefinition": False,
    ...                 "owner": "Refinitiv",
    ...             },
    ...             "forwardCurves": [
    ...                 {
    ...                     "curvePoints": [
    ...                         {
    ...                             "endDate": "2021-02-01",
    ...                             "startDate": "2021-02-01",
    ...                             "discountFactor": 1.0,
    ...                             "ratePercent": 7.040811073443143,
    ...                             "tenor": "0D",
    ...                         },
    ...                         {
    ...                             "endDate": "2021-02-04",
    ...                             "startDate": "2021-02-01",
    ...                             "discountFactor": 0.999442450671571,
    ...                             "ratePercent": 7.040811073443143,
    ...                             "tenor": "1D",
    ...                         },
    ...                     ],
    ...                     "forwardCurveTag": "ForwardTag",
    ...                     "forwardStart": "2021-02-01",
    ...                     "indexTenor": "3M",
    ...                 }
    ...             ],
    ...         },
    ...     ]
    ... }
    """
    datas = raw.get("data", [])
    datas = datas or []
    dfs = []
    for data in datas:

        error = data.get("error")
        if error:
            continue

        forward_curves = data.get("forwardCurves")

        for forward_curve in forward_curves:
            curve_points = forward_curve.get("curvePoints")

            d = {}
            for curve_point in curve_points:
                for key, value in curve_point.items():
                    values = d.setdefault(key, [])
                    values.append(value)

            df = pd.DataFrame(d)
            df = df.convert_dtypes()
            dfs.append(df)

    df = pd.concat(dfs, ignore_index=True)
    df = convert_df_columns_to_datetime(df, "Date", utc=True, delete_tz=True)
    return df


def zc_curve_definitions_build_df(raw, **kwargs):
    data = raw.get("data", [])
    data = data or []
    curve_definitions = [d for d in data if d for d in d.get("curveDefinitions")]
    df = pd.DataFrame(curve_definitions)

    if not df.empty:
        df = df.convert_dtypes()
        df = convert_df_columns_to_datetime(df, "Date", utc=True, delete_tz=True)
    return df


class OneCurveData(Data):
    def __init__(self, raw, create_curves, dfbuilder):
        Data.__init__(self, raw, dfbuilder=dfbuilder)
        self._create_curves = create_curves
        self._curve = None

    @property
    def curve(self) -> Curve:
        if self._curve is None:
            curve = self._create_curves(self._raw)
            self._curve = curve[0]
        return self._curve


class CurvesData(Data):
    def __init__(self, raw, create_curves, dfbuilder):
        Data.__init__(self, raw, dfbuilder=dfbuilder)
        self._create_curves = create_curves
        self._curves = None

    @property
    def curves(self) -> List[Curve]:
        if self._curves is None:
            self._curves = self._create_curves(self._raw)
        return self._curves


def make_create_forward_curves(x_axis: str, y_axis: str) -> Callable:
    """
    Parameters
    ----------
    x_axis: str
        Name of key in curve point data for build X axis
    y_axis: str
        Name of key in curve point data for build Y axis

    Returns
    -------
    Callable
    """

    def create_forward_curves(raw: dict) -> list:
        """
        Curve point in "curvePoints":
        {
            "discountFactor": 1.0,
            "endDate": "2021-02-01",
            "ratePercent": -2.330761285491212,
            "startDate": "2021-02-01",
            "tenor": "0D"
        }
        Parameters
        ----------
        raw

        Returns
        -------
        list of ForwardCurve
        """
        curves = []
        for data in raw.get("data", []):
            for forward_curve in data.get("forwardCurves", []):
                x, y = [], []
                for point in forward_curve.get("curvePoints"):
                    end_date = point.get(x_axis)
                    x.append(end_date)
                    discount_factor = point.get(y_axis)
                    y.append(discount_factor)

                x = np.array(x, dtype=get_type_by_axis(x_axis))
                y = np.array(y, dtype=get_type_by_axis(y_axis))
                curve = ForwardCurve(x, y, **forward_curve)
                curves.append(curve)

        return curves

    return create_forward_curves


def make_create_zc_curves(x_axis: str, y_axis: str) -> Callable:
    """
    Parameters
    ----------
    x_axis: str
        Name of key in curve point data for build X axis
    y_axis: str
        Name of key in curve point data for build Y axis

    Returns
    -------
    Callable
    """

    def create_zc_curves(raw: dict) -> list:
        """
        Curve point in "curvePoints":
        {
            "discountFactor": 1.0,
            "endDate": "2021-07-27",
            "ratePercent": -0.7359148312458879,
            "startDate": "2021-07-27",
            "tenor": "ON",
            "instruments": [
                {
                    "instrumentCode": "SARON.S"
                }
            ]
        }
        Parameters
        ----------
        raw

        Returns
        -------
        list of ZcCurve
        """
        curves = []
        for data in raw.get("data", []):
            for index_tenor, zc_curve in data.get("curves", {}).items():
                x, y = [], []
                for point in zc_curve.get("curvePoints"):
                    end_date = point.get(x_axis)
                    x.append(end_date)
                    discount_factor = point.get(y_axis)
                    y.append(discount_factor)

                x = np.array(x, dtype=get_type_by_axis(x_axis))
                y = np.array(y, dtype=get_type_by_axis(y_axis))
                curve = ZcCurve(x, y, index_tenor, **zc_curve)
                curves.append(curve)

        return curves

    return create_zc_curves


curves_maker_by_content_type = {
    ContentType.FORWARD_CURVE: make_create_forward_curves(
        x_axis="endDate", y_axis="discountFactor"
    ),
    ContentType.ZC_CURVES: make_create_zc_curves(
        x_axis="endDate", y_axis="discountFactor"
    ),
}


def get_curves_maker(content_type):
    curves_maker = curves_maker_by_content_type.get(content_type)

    if not curves_maker:
        raise ValueError(f"Cannot find curves_maker for content_type={content_type}")

    return curves_maker


# ---------------------------------------------------------------------------
#   Response factory
# ---------------------------------------------------------------------------


class CurvesResponseFactory(ResponseFactory):
    def create_success(self, *args, **kwargs):
        data = args[0]
        raw = data.get("content_data")
        universe = kwargs.get("universe")
        inst = self.response_class(is_success=True, **data)
        dfbuilder = self.get_dfbuilder(kwargs)
        content_type = kwargs.get("__content_type__")

        if content_type is ContentType.ZC_CURVE_DEFINITIONS:
            data = Data(raw, dfbuilder=dfbuilder)

        else:
            curves_maker = get_curves_maker(content_type)
            if iterable(universe):
                data = CurvesData(raw, curves_maker, dfbuilder)

            else:
                data = OneCurveData(raw, curves_maker, dfbuilder)

        inst.data = data
        inst.data._owner = inst
        return inst


# ---------------------------------------------------------------------------
#   Data provider
# ---------------------------------------------------------------------------

curves_data_provider = DataProvider(
    request=CurvesAndSurfacesRequestFactory(),
    response=CurvesResponseFactory(),
    validator=ValidatorContainer(content_validator=CurvesContentValidator()),
)



# coding: utf8
# contract_gen 2020-05-18 08:30:59.207816

__all__ = ["StubRule"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class StubRule(Enum):
    """
    The rule that defines whether coupon roll dates are aligned on the  maturity or
       the issue date.

    The possible values are:
       - ShortFirstProRata (to create a short period between the start date and the
         first coupon date, and pay a smaller amount of interest for the short
         period.All coupon dates are calculated backward from the maturity date),
       - ShortFirstFull (to create a short period between the start date and the first
         coupon date, and pay a regular coupon on the first coupon date. All coupon
         dates are calculated backward from the maturity date),
       - LongFirstFull (to create a long period between the start date and the second
         coupon date, and pay a regular coupon on the second coupon date. All coupon
         dates are calculated backward from the maturity date),
       - ShortLastProRata (to create a short period between the last payment date and
         maturity, and pay a smaller amount of interest for the short period. All
         coupon dates are calculated forward from the start date). This property may
         also be used in conjunction with first_regular_payment_date and
         last_regular_payment_date; in that case the following values can be defined,
       - Issue (all dates are aligned on the issue date),
       - Maturity (all dates are aligned on the maturity date).
    """

    ISSUE = "Issue"
    LONG_FIRST_FULL = "LongFirstFull"
    MATURITY = "Maturity"
    SHORT_FIRST_FULL = "ShortFirstFull"
    SHORT_FIRST_PRO_RATA = "ShortFirstProRata"
    SHORT_LAST_PRO_RATA = "ShortLastProRata"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(StubRule, _STUB_RULE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_STUB_RULE_VALUES_IN_LOWER_BY_STUB_RULE, some)


_STUB_RULE_VALUES = tuple(t.value for t in StubRule)
_STUB_RULE_VALUES_IN_LOWER_BY_STUB_RULE = {
    name.lower(): item for name, item in StubRule.__members__.items()
}



# coding: utf8
# contract_gen 2020-09-02 06:21:29.297010

__all__ = ["QuoteFallbackLogic"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class QuoteFallbackLogic(Enum):
    """
    Enumeration used to define the fallback logic for the quotation of the instrument.

    Available values are:
    - "None": it means that there's no fallback logic. For example, if the user asks
      for a "Ask" price and instrument is only quoted with a "Bid" price, it is an
      error case.
    - "BestField" : it means that there's a fallback logic to use another market
      data field as quoted price. For example, if the user asks for a "Ask" price
      and instrument is only quoted with a "Bid" price, "Bid" price can be used.
    """

    BEST_FIELD = "BestField"
    NONE = "None"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(QuoteFallbackLogic, _QUOTE_FALLBACK_LOGIC_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _QUOTE_FALLBACK_LOGIC_VALUES_IN_LOWER_BY_QUOTE_FALLBACK_LOGIC, some
        )


_QUOTE_FALLBACK_LOGIC_VALUES = tuple(t.value for t in QuoteFallbackLogic)
_QUOTE_FALLBACK_LOGIC_VALUES_IN_LOWER_BY_QUOTE_FALLBACK_LOGIC = {
    name.lower(): item for name, item in QuoteFallbackLogic.__members__.items()
}



# coding: utf8
# contract_gen 2020-06-15 10:07:47.554306

__all__ = ["InterpolationMode"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class InterpolationMode(Enum):
    CUBIC_DISCOUNT = "CubicDiscount"
    CUBIC_RATE = "CubicRate"
    CUBIC_SPLINE = "CubicSpline"
    FORWARD_MONOTONE_CONVEX = "ForwardMonotoneConvex"
    LINEAR = "Linear"
    LOG = "Log"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(InterpolationMode, _INTERPOLATION_MODE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(
            _INTERPOLATION_MODE_VALUES_IN_LOWER_BY_INTERPOLATION_MODE, some
        )


_INTERPOLATION_MODE_VALUES = tuple(t.value for t in InterpolationMode)
_INTERPOLATION_MODE_VALUES_IN_LOWER_BY_INTERPOLATION_MODE = {
    name.lower(): item for name, item in InterpolationMode.__members__.items()
}



# coding: utf8

__all__ = ["FxLegType"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class FxLegType(Enum):
    """
    The enumeration that specifies the type of the leg :
        'Spot',
        'FxForward',
        'FxNonDeliverableForward',
        'SwapNear' or 'SwapFar'.
    Mandatory for MultiLeg, FwdFwdSwap, or Swap contracts.
    Optional for Spot and Forwards contracts.
    """

    FX_FORWARD = "FxForward"
    FX_NON_DELIVERABLE_FORWARD = "FxNonDeliverableForward"
    FX_SPOT = "FxSpot"
    SWAP_FAR = "SwapFar"
    SWAP_NEAR = "SwapNear"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(FxLegType, _FX_LEG_TYPE_VALUES, some)

    @staticmethod
    def normalize(some):
        return _normalize(_FX_LEG_TYPE_VALUES_IN_LOWER_BY_FX_LEG_TYPE, some)


_FX_LEG_TYPE_VALUES = tuple(t.value for t in FxLegType)
_FX_LEG_TYPE_VALUES_IN_LOWER_BY_FX_LEG_TYPE = {
    name.lower(): item for name, item in FxLegType.__members__.items()
}



from typing import Optional, TYPE_CHECKING

from .._config_functions import load_config, get_config
from .._core.session import is_open, is_closed
from .._core.session import get_default, set_default
from .._core.session._session_definition import Definition
from .._log import root_logger


if TYPE_CHECKING:
    from .._core.session import Session


def open_session(
    name: Optional[str] = None,
    app_key: Optional[str] = None,
    config_name: Optional[str] = None,
) -> "Session":
    """
    Opens session.
    Without params opens default session from default config

    Parameters
    ----------
        name: str,
            if provided opens named session from config,
            It should be in the following format: 'session_type.session_name'
        app_key: str,
            if provided overrides app_key
        config_name: str,
            if provided overrides config and set it as default

    Returns
    -------
    Session
    """
    # Because it's a path in configuration profile, not a session name,
    # but we can't change argument name, because it's a public API
    config_path = name
    del name

    _definition = Definition
    _load_config = load_config
    _config_object = get_config()
    _set_default = set_default
    return _open_session(**locals())


def close_session() -> None:
    """
    Close default session

    Returns
    -------
    None
    """
    _get_default = get_default
    _close_session(**locals())


def _open_session(**kwargs) -> "Session":
    config_name_for_err = "default"
    app_key = kwargs.pop("app_key")
    config_name = kwargs.pop("config_name")

    _definition = kwargs.pop("_definition")
    _load_config = kwargs.pop("_load_config")
    _config_object = kwargs.pop("_config_object")
    _set_default = kwargs.pop("_set_default")

    if config_name:
        _load_config(config_name)
        config_name_for_err = "config_name"

    name = kwargs.pop("config_path") or _config_object.get_param("sessions.default")

    try:
        _config_object.get_param(f"sessions.{name}")
    except Exception:
        raise NameError(
            f"Cannot open session {name}\n"
            f"This session is not defined in the {config_name_for_err} configuration file"
        )

    if app_key:
        _config_object.set_param(param=f"sessions.{name}.app-key", value=app_key)

    session = _definition(name=name).get_session()

    _set_default(session)
    session.open()

    return session


def _close_session(**kwargs) -> None:
    _get_default = kwargs.pop("_get_default")

    try:
        default_session = _get_default()
    except AttributeError:
        root_logger().info("Nо default session to close")
    else:
        if is_open(default_session):
            default_session.info("Closing session")
            default_session.close()
        if is_closed(default_session):
            default_session.info("Session is closed")



# coding: utf8

__all__ = ["IndexSpreadCompoundingMethod"]

from enum import Enum, unique
from ._common_tools import _convert_to_str, _normalize


@unique
class IndexSpreadCompoundingMethod(Enum):
    ISDA_COMPOUNDING = "IsdaCompounding"
    ISDA_FLAT_COMPOUNDING = "IsdaFlatCompounding"
    NO_COMPOUNDING = "NoCompounding"

    @staticmethod
    def convert_to_str(some):
        return _convert_to_str(
            IndexSpreadCompoundingMethod, _INDEX_SPREAD_COMPOUNDING_METHOD_VALUES, some
        )

    @staticmethod
    def normalize(some):
        return _normalize(
            _INDEX_SPREAD_COMPOUNDING_METHOD_VALUES_IN_LOWER_BY_INDEX_SPREAD_COMPOUNDING_METHOD,
            some,
        )


_INDEX_SPREAD_COMPOUNDING_METHOD_VALUES = tuple(
    t.value for t in IndexSpreadCompoundingMethod
)
_INDEX_SPREAD_COMPOUNDING_METHOD_VALUES_IN_LOWER_BY_INDEX_SPREAD_COMPOUNDING_METHOD = {
    name.lower(): item
    for name, item in IndexSpreadCompoundingMethod.__members__.items()
}
